#!/usr/bin/env python
"""
argmax, argmin and argsort for array_like and Python iterables.

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2014-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   argmax
   argmin
   argsort

History
    * Written Dec 2014 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added argmin, argmax, Jul 2019, Matthias Cuntz
    * Using numpy docstring format, extending examples from numpy docstrings,
      May 2020, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
import numpy as np


__all__ = ['argmax', 'argmin', 'argsort']


def argmax(a, *args, **kwargs):
    """
    Wrapper for numpy.argmax, numpy.ma.argmax, and using max for Python
    iterables.

    Passes all keywords directly to the individual routines, i.e.

    .. code-block:: python

       numpy.argmax(a, axis=None, out=None)
       numpy.ma.argmax(self, axis=None, fill_value=None, out=None)

    No keyword will be passed to the max routine for Python iterables.

    Parameters
    ----------
    a : array_like
        input array, masked array, or Python iterable
    *args : optional
        all arguments of numpy.argmax or numpy.ma.argmax
    **kwargs : optional
        all keyword arguments of numpy.argmax or numpy.ma.argmax

    Returns
    -------
    index_array : ndarray, int
        Array of indices of the largest element in input array `a`. It has the
        same shape as `a.shape` with the dimension along `axis` removed.
        `a[np.unravel_index(argmax(a), a.shape)]` is the maximum value of `a`.

    Notes
    -----
    argmax for iterables was taken from
    https://stackoverflow.com/questions/16945518/finding-the-index-of-the-value-which-is-the-min-or-max-in-python

    Examples
    --------
    One-dimensional array

    >>> import numpy as np
    >>> a = np.array([0,4,6,2,1,5,3,5])
    >>> ii = argmax(a)
    >>> print(ii)
    2
    >>> print(a[ii])
    6

    One-dimensional masked array

    >>> a = np.ma.array([0,4,6,2,1,5,3,5], mask=[0,0,1,1,0,0,0,0])
    >>> ii = argmax(a)
    >>> print(ii)
    5
    >>> print(a[ii])
    5
    >>> ii = argmax(a, fill_value=6)
    >>> print(ii)
    2

    List

    >>> a = [0,4,6,2,1,5,3,5]
    >>> ii = argmax(a)
    >>> print(ii)
    2
    >>> print(a[ii])
    6

    Examples from numpy.argmax docstring

    >>> a = np.arange(6).reshape(2,3) + 10
    >>> a
    array([[10, 11, 12],
           [13, 14, 15]])
    >>> argmax(a)
    5
    >>> argmax(a, axis=0)
    array([1, 1, 1])
    >>> argmax(a, axis=1)
    array([2, 2])

    >>> # Indexes of the maximal elements of a N-dimensional array:
    >>> ind = np.unravel_index(np.argmax(a, axis=None), a.shape)
    >>> ind
    (1, 2)
    >>> a[ind]
    15

    >>> b = np.arange(6)
    >>> b[1] = 5
    >>> b
    array([0, 5, 2, 3, 4, 5])
    >>> argmax(b)  # Only the first occurrence is returned.
    1

    """
    if isinstance(a, np.ma.MaskedArray):
        return np.ma.argmax(a, *args, **kwargs)
    elif isinstance(a, np.ndarray):
        return np.argmax(a, *args, **kwargs)
    else:
        return _argmax(a)


def argmin(a, *args, **kwargs):
    """
    Wrapper for numpy.argmin, numpy.ma.argmin, and using min for Python
    iterables

    Passes all keywords directly to the individual routines, i.e.

    .. code-block:: python

       numpy.argmin(a, axis=None, out=None)
       numpy.ma.argmin(self, axis=None, fill_value=None, out=None)

    No keyword will be passed to the min routine for Python iterables.

    Parameters
    ----------
    a : array_like
        input array, masked array, or Python iterable
    *args : optional
        all arguments of numpy.argmin or numpy.ma.argmin
    **kwargs : optional
        all keyword arguments of numpy.argmin or numpy.ma.argmin

    Returns
    -------
    index_array : ndarray, int
        Array of indices of the largest element in input array `a`. It has the
        same shape as `a.shape` with the dimension along `axis` removed.
        `a[np.unravel_index(argmin(a), a.shape)]` is the minimum value of `a`.

    Notes
    -----
    argmin for iterables was taken from
    https://stackoverflow.com/questions/16945518/finding-the-index-of-the-value-which-is-the-min-or-max-in-python

    Examples
    --------
    One-dimensional array

    >>> import numpy as np
    >>> a = np.array([0,4,6,2,1,5,3,5])
    >>> ii = argmin(a)
    >>> print(ii)
    0
    >>> print(a[ii])
    0

    One-dimensional masked array

    >>> a = np.ma.array([0,4,6,2,1,5,3,5], mask=[1,0,1,1,0,0,0,0])
    >>> ii = argmin(a)
    >>> print(ii)
    4
    >>> print(a[ii])
    1
    >>> ii = argmin(a, fill_value=1)
    >>> print(ii)
    0

    List

    >>> a = [0,4,6,2,1,5,3,5]
    >>> ii = argmin(a)
    >>> print(ii)
    0
    >>> print(a[ii])
    0

    Examples from numpy.argmin docstring

    >>> a = np.arange(6).reshape(2,3) + 10
    >>> a
    array([[10, 11, 12],
           [13, 14, 15]])
    >>> argmin(a)
    0
    >>> argmin(a, axis=0)
    array([0, 0, 0])
    >>> argmin(a, axis=1)
    array([0, 0])

    >>> # Indices of the minimum elements of a N-dimensional array:
    >>> ind = np.unravel_index(argmin(a, axis=None), a.shape)
    >>> ind
    (0, 0)
    >>> a[ind]
    10

    >>> b = np.arange(6) + 10
    >>> b[4] = 10
    >>> b
    array([10, 11, 12, 13, 10, 15])
    >>> argmin(b)  # Only the first occurrence is returned.
    0

    """
    if isinstance(a, np.ma.MaskedArray):
        return np.ma.argmin(a, *args, **kwargs)
    elif isinstance(a, np.ndarray):
        return np.argmin(a, *args, **kwargs)
    else:
        return _argmin(a)


def argsort(a, *args, **kwargs):
    """
    Wrapper for numpy.argsort, numpy.ma.argsort, and using sorted for Python
    iterables

    Passes all keywords directly to the individual routines, i.e.

    .. code-block:: python

       numpy.argsort(a, axis=-1, kind='quicksort', order=None)
       numpy.ma.argsort(a, axis=None, kind='quicksort', order=None,
                        fill_value=None)
       sorted(iterable[, cmp[, key[, reverse]]])

    Only key cannot be given for Python iterables because the input array is
    used as key in the sorted function.

    Parameters
    ----------
    a : array_like
        input array, masked array, or Python iterable
    *args : optional
        all arguments of numpy.argsort, numpy.ma.argsort, and sorted (except
        key argument)
    **kwargs : optional
        all keyword arguments of numpy.argsort, numpy.ma.argsort, and sorted
        (except key argument)

    Returns
    -------
    index_array : ndarray, int
        Array of indices that sort `a` along the specified `axis`.
        If `a` is one-dimensional, `a[index_array]` yields a sorted `a`.

    Notes
    -----
    argsort for iterables was taken from
    http://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python
    http://stackoverflow.com/questions/3071415/efficient-method-to-calculate-the-rank-vector-of-a-list-in-python

    Examples
    --------
    1D array

    >>> import numpy as np
    >>> a = np.array([0,4,6,2,1,5,3,5])
    >>> ii = argsort(a)
    >>> print(a[ii])
    [0 1 2 3 4 5 5 6]

    >>> ii = argsort(a, kind='quicksort')
    >>> print(a[ii])
    [0 1 2 3 4 5 5 6]

    1D masked array

    >>> a = np.ma.array([0,4,6,2,1,5,3,5], mask=[0,0,1,1,0,0,0,0])
    >>> ii = argsort(a)
    >>> print(a[ii])
    [0 1 3 4 5 5 -- --]

    >>> ii = argsort(a, fill_value=1)
    >>> print(a[ii])
    [0 -- -- 1 3 4 5 5]

    list

    >>> a = [0,4,6,2,1,5,3,5]
    >>> ii = argsort(a)
    >>> b = [ a[i] for i in ii ]
    >>> print(b)
    [0, 1, 2, 3, 4, 5, 5, 6]

    >>> a = [0,4,6,2,1,5,3,5]
    >>> ii = argsort(a, reverse=True)
    >>> b = [ a[i] for i in ii ]
    >>> print(b)
    [6, 5, 5, 4, 3, 2, 1, 0]


    Examples from numpy.argsort docstring

    >>> # One-dimensional array:
    >>> x = np.array([3, 1, 2])
    >>> argsort(x)
    array([1, 2, 0])

    >>> # Two-dimensional array:
    >>> x = np.array([[0, 3], [2, 2]])
    >>> x
    array([[0, 3],
           [2, 2]])
    >>> ind = argsort(x, axis=0)  # sorts along first axis (down)
    >>> ind
    array([[0, 1],
           [1, 0]])
    >>> np.take_along_axis(x, ind, axis=0)  # same as np.sort(x, axis=0)
    array([[0, 2],
           [2, 3]])
    >>> ind = argsort(x, axis=1)  # sorts along last axis (across)
    >>> ind
    array([[0, 1],
           [0, 1]])
    >>> np.take_along_axis(x, ind, axis=1)  # same as np.sort(x, axis=1)
    array([[0, 3],
           [2, 2]])

    >>> # Indices of the sorted elements of a N-dimensional array:
    >>> ind = np.unravel_index(argsort(x, axis=None), x.shape)
    >>> ind
    (array([0, 1, 1, 0]), array([0, 0, 1, 1]))
    >>> x[ind]  # same as np.sort(x, axis=None)
    array([0, 2, 2, 3])

    >>> # Sorting with keys:
    >>> x = np.array([(1, 0), (0, 1)], dtype=[('x', '<i4'), ('y', '<i4')])
    >>> x
    array([(1, 0), (0, 1)],
          dtype=[('x', '<i4'), ('y', '<i4')])
    >>> argsort(x, order=('x','y'))
    array([1, 0])
    >>> argsort(x, order=('y','x'))
    array([0, 1])

    """
    if isinstance(a, np.ma.MaskedArray):
        return np.ma.argsort(a, *args, **kwargs)
    elif isinstance(a, np.ndarray):
        return np.argsort(a, *args, **kwargs)
    else:
        return _argsort(a, *args, **kwargs)


# same as numpy.argmax but for python iterables
def _argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]


# same as numpy.argmin but for python iterables
def _argmin(iterable):
    return min(enumerate(iterable), key=lambda x: x[1])[0]


# same as numpy.argsort but for python iterables
def _argsort(seq, *args, **kwargs):
    if 'key' in kwargs:
        raise KeyError('keyword key cannot be given to argsort.')
    return sorted(range(len(seq)), *args, key=seq.__getitem__, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
Function for Morris' method of Elementary Effects.

This function was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2017-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   ee
   screening

History
    * Written Dec 2017 by Matthias Cuntz (mc (at) macu (dot) de)
    * Output also (npara,3)-array for nt=1, Dec 2017, Matthias Cuntz
    * Removed call to external programs: use exe_wrappers,
      Jan 2018, Matthias Cuntz
    * Function can return multiple outputs, e.g. time series,
      Jan 2018, Matthias Cuntz
    * Python 3: map returns iterator, so list(map), Jul 2018, Fabio Gennaretti
    * Bug: default ntotal was not set if ntotal<0 (but nt instead),
      Dec 2019, Matthias Cuntz
    * Make numpy docstring format, Dec 2019, Matthias Cuntz
    * x0 optional, Jan 2020, Matthias Cuntz
    * Distinguish iterable and array_like parameter types;
      added seed keyword to screening/ee, Jan 2020, Matthias Cuntz
    * InputError does not exist, use TypeError, Feb 2020, Matthias Cuntz
    * Use new names of kwargs of morris_sampling and elementary_effects,
      Feb 2020, Matthias Cuntz
    * Make number of final trajectories an argument instead of a keyword
      argument, Feb 2020, Matthias Cuntz
    * Sample not only from uniform distribution but allow all distributions of
      scipy.stats, Mar 2020, Matthias Cuntz
    * Code refactoring, Sep 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
from __future__ import division, absolute_import, print_function
import numpy as np
from pyjams import morris_sampling, elementary_effects


__all__ = ['screening', 'ee']


def screening(func, lb, ub, nt, x0=None, mask=None,
              nsteps=6, ntotal=None,
              dist=None, distparam=None,
              seed=None,
              processes=1, pool=None,
              verbose=0):
    """
    Parameter screening using Morris' method of Elementary Effects.

    Note, the input function must be callable as `func(x)`.

    Parameters
    ----------
    func : callable
        Python function callable as `func(x)` with the function parameters *x*.
    lb : array_like
        Lower bounds of parameters
        or lower fraction of percent point function ppf if distribution given.
        Be aware that the percent point function *ppf* of most continuous
        distributions is infinite at 0.
    ub : array_like
        Upper bounds of parameters
        or upper fraction of percent point function ppf if distribution given.
        Be aware that the percent point function *ppf* of most continuous
        distributions is infinite at 1.
    nt : int
        Number of trajectories used for screening.
    x0 : array_like, optional
        Parameter values used with `mask==0`.
    mask : array_like, optional
        Include (1,True) or exclude (0,False) parameters in screening (default:
        include all parameters).
    nsteps : int, optional
        Number of steps along one trajectory (default: 6)
    ntotal : int, optional
        Total number of sampled trajectories to select the nt most different
        trajectories.
        If None: `max(nt**2,10*nt)` (default: None)
    dist : list, optional
        List of None or scipy.stats distribution objects for each factor having
        the method ppf, Percent Point Function (Inverse of CDF) (default: None).
        If None, the uniform distribution will be sampled from lower bound `lb`
        to upper bound `ub`.
        If dist is scipy.stats.uniform, the ppf will be sampled from the lower
        fraction given in `lb` and the upper fraction in `ub`. The sampling
        interval is then given by the parameters `loc=lower` and
        `scale=interval=upper-lower` in distparam. This means
        `dist=None`, `lb=a`, `ub=b`
        corresponds to
        `lb=0`, `ub=1`, `dist=scipy.stats.uniform`, `distparam=[a,b-a]`
    distparam : list, optional
        List with tuples with parameters as required for `dist` (default:
        (0, 1)).
        All distributions of scipy.stats have location and scale parameters, at
        least. `loc` and `scale` are implemented as keyword arguments in
        scipy.stats. Other parameters such as the shape parameter of the gamma
        distribution must hence be given first, e.g. `(shape,loc,scale)` for
        the gamma distribution.
        `distparam` is ignored if `dist` is None.
        The percent point function *ppf* is called like this:
        `dist(*distparam).ppf(x)`
    seed : int or array_like
        Seed for numpy's random number generator (default: None).
    processes : int, optional
        The number of processes to use to evaluate objective function and
        constraints (default: 1).
    pool : `schwimmbad` pool object, optional
        Generic map function used from module `schwimmbad
        <https://schwimmbad.readthedocs.io/en/latest/>`_, which provides,
        serial, multiprocessor, and MPI mapping functions (default: None).
        The pool will be chosen automatically if `pool` is None. It is chosen
        with:

        .. code-block:: python

           schwimmbad.choose_pool(mpi=True/False, processes=processes).

        MPI pools can only be opened and closed once. If you want to use
        screening several times in one program, then you have to choose the
        pool, pass it to screening, and later close the pool in the calling
        program.

    verbose : int, optional
        Print progress report during execution if `verbose>0` (default: 0).

    Returns
    -------
    (nparameter, 3) ndarray
        2D-array with 3 entries per parameters:

            1. mean of absolute elementary effects over all nt trajectories
               (`mu*`)
            2. mean of elementary effects over all nt trajectories (`mu`)
            3. standard deviation of elementary effects over all
               nt trajectories (`sigma`) or zero if `nt==1`

    Examples
    --------
    >>> from functools import partial
    >>> import numpy as np
    >>> from partialwrap import function_wrapper
    >>> from pyjams.functions import fmorris
    >>> seed = 1023
    >>> np.random.seed(seed=seed)
    >>> npars = 20
    >>> beta0              = 0.
    >>> beta1              = np.random.standard_normal(npars)
    >>> beta1[:10]         = 20.
    >>> beta2              = np.random.standard_normal((npars,npars))
    >>> beta2[:6,:6]       = -15.
    >>> beta3              = np.zeros((npars,npars,npars))
    >>> beta3[:5,:5,:5]    = -10.
    >>> beta4              = np.zeros((npars,npars,npars,npars))
    >>> beta4[:4,:4,:4,:4] = 5.
    >>> arg   = [beta0, beta1, beta2, beta3, beta4]
    >>> kwarg = {}
    >>> func  = partial(function_wrapper, fmorris, arg, kwarg)
    >>> lb    = np.zeros(npars)
    >>> ub    = np.ones(npars)
    >>> nt      = 20
    >>> ntotal  = 10*nt
    >>> nsteps  = 6
    >>> verbose = 0
    >>> out = screening(func, lb, ub, nt, x0=None, mask=None,
    ...                 nsteps=nsteps, ntotal=ntotal, processes=4,
    ...                 verbose=verbose)
    >>> print(out[0:3,0])
    [60.7012889  67.33372626 48.46673528]

    """
    # Get MPI communicator
    try:  # pragma: no cover
        from mpi4py import MPI
        comm  = MPI.COMM_WORLD
        csize = comm.Get_size()
        crank = comm.Get_rank()
    except ImportError:
        comm  = None
        csize = 1
        crank = 0

    # Checks
    # Bounds
    assert len(lb) == len(ub), (
        'Lower- and upper-bounds must have the same lengths.')
    lb = np.array(lb)
    ub = np.array(ub)
    # Mask
    if mask is None:
        assert np.all(ub >= lb), (
            'All upper-bounds must be greater or equal than lower-bounds.')
    else:
        assert len(mask) == len(ub), (
            'Mask and bounds must have the same lengths.')
        if x0 is None:
            raise TypeError('x0 must be given if mask is set')
        x0 = np.array(x0)
        if not np.all(mask):
            assert len(mask) == len(x0), (
                'Mask and x0 must have the same lengths.')
        assert np.all(ub[mask] >= lb[mask]), (
            'All unmasked upper-bounds must be greater or equal than'
            ' lower-bounds.')

    # Set defaults
    npara = len(lb)
    if mask is None:
        imask  = np.ones(npara, dtype=bool)
    else:
        imask  = mask
    nmask = np.sum(imask)
    if ntotal is None:
        ntotal = max(nt**2, 10*nt)

    # Seed random number generator
    if seed is not None:  # pragma: no cover
        # same on all ranks because trajectories are sampled on all ranks
        np.random.seed(seed=seed)

    # Sample trajectories
    if (crank == 0) and (verbose > 0):
        print('Sample trajectories')
    if dist is None:
        idist      = None
        idistparam = None
    else:
        idist      = [ dist[i]      for i in np.where(imask)[0] ]
        idistparam = [ distparam[i] for i in np.where(imask)[0] ]
    tmatrix, tvec = morris_sampling(nmask, lb[imask], ub[imask], nt,
                                    nsteps=nsteps, ntotal=ntotal,
                                    dist=idist, distparam=idistparam,
                                    Diagnostic=False)

    if mask is None:
        x = tmatrix
    else:
        # Set input vector to trajectories and masked elements = x0
        x = np.tile(x0, tvec.size).reshape(tvec.size, npara)  # default to x0
        x[:, imask] = tmatrix  # replaced unmasked with trajectorie values

    # Choose the right mapping function: single, multiprocessor or mpi
    if pool is None:
        import schwimmbad
        ipool = schwimmbad.choose_pool(mpi=False if csize == 1 else True,
                                       processes=processes)
    else:
        ipool = pool

    # Calculate all model runs
    if (crank == 0) and (verbose > 0):
        print('Calculate objective functions')
    fx = np.array(list(ipool.map(func, x)))

    # Calc elementary effects
    if (crank == 0) and (verbose > 0):
        print('Calculate elementary effects')
    if fx.ndim == 1:
        fx  = fx[:, np.newaxis]
        nfx = 1
    else:
        nfx = fx.shape[1]
    out = np.zeros((nfx, npara, 3))
    for j in range(nfx):
        sa, res = elementary_effects(nmask, tmatrix, tvec, fx[:, j],
                                     nsteps=nsteps, Diagnostic=False)
        # Output with zero for all masked parameters
        if nt == 1:
            out[j, imask, 0] = np.abs(sa[:, 0])
            out[j, imask, 1] = sa[:, 0]
        else:
            out[j, imask, :] = res

    if nfx == 1:
        out = out[0, :, :]

    if pool is None:
        ipool.close()

    return out


def ee(*args, **kwargs):
    """
    Wrapper function for :func:`screening`.
    """
    return screening(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import schwimmbad
    # import jams

    # import time as ptime
    # t1 = ptime.time()

    # try:
    #     from mpi4py import MPI
    #     comm  = MPI.COMM_WORLD
    #     csize = comm.Get_size()
    #     crank = comm.Get_rank()
    # except ImportError:
    #     comm  = None
    #     csize = 1
    #     crank = 0

    # seed = 1025

    # if seed is not None:
    #     np.random.seed(seed=seed)

    # if False:
    #     # work with global parameters
    #     def morris(X):
    #         return jams.functions.morris(
    #             X, beta0, beta1, beta2, beta3, beta4)
    #     func = morris
    #     npars = 20
    #     beta0              = 0.
    #     beta1              = np.random.standard_normal(npars)
    #     beta1[:10]         = 20.
    #     beta2              = np.random.standard_normal((npars,npars))
    #     beta2[:6,:6]       = -15.
    #     beta3              = np.zeros((npars,npars,npars))
    #     beta3[:5,:5,:5]    = -10.
    #     beta4              = np.zeros((npars,npars,npars,npars))
    #     beta4[:4,:4,:4,:4] = 5.
    #     lb    = np.zeros(npars)
    #     ub    = np.ones(npars)
    #     nt      = 20
    #     ntotal  = 10*nt
    #     nsteps  = 6
    #     verbose = 1

    #     out = screening(func, lb, ub, nt, x0=None, mask=None, nsteps=nsteps,
    #                     ntotal=ntotal, processes=4, verbose=1)

    #     t2    = ptime.time()

    #     if crank == 0:
    #         strin = ('[m]: {:.1f}'.format(
    #             (t2-t1)/60.)
    #             if (t2-t1)>60. else '[s]: {:d}'.format(int(t2-t1)))
    #         print('Time elapsed: ', strin)
    #         print('mu*, mu, std: ', out)
    #         out1 = out / out.max(axis=0)
    #         mustar = out1[:,0]
    #         mustar = np.sort(mustar)
    #         import matplotlib.pyplot as plt
    #         plt.plot(mustar, 'ro')
    #         plt.show()

    # if False:
    #     # pass parameters
    #     func = jams.functions.morris
    #     npars = 20
    #     lb    = np.zeros(npars)
    #     ub    = np.ones(npars)
    #     beta0              = 0.
    #     beta1              = np.random.standard_normal(npars)
    #     beta1[:10]         = 20.
    #     beta2              = np.random.standard_normal((npars,npars))
    #     beta2[:6,:6]       = -15.
    #     beta3              = np.zeros((npars,npars,npars))
    #     beta3[:5,:5,:5]    = -10.
    #     beta4              = np.zeros((npars,npars,npars,npars))
    #     beta4[:4,:4,:4,:4] = 5.
    #     args = [beta0, beta1, beta2, beta3, beta4] # Morris
    #     nt = 20
    #     ntotal  = 10*nt
    #     nsteps = 6
    #     verbose = 1

    #     out = screening(func, lb, ub, nt, arg=args, mask=None, nsteps=nsteps,
    #                     ntotal=ntotal, processes=4, verbose=1)
    #     t2    = ptime.time()

    #     if crank == 0:
    #         strin = ('[m]: {:.1f}'.format(
    #                  (t2-t1)/60.)
    #                  if (t2-t1)>60. else '[s]: {:d}'.format(int(t2-t1)))
    #         print('Time elapsed: ', strin)
    #         print('mu*, mu, std: ', out)
    #         out1 = out / out.max(axis=0)
    #         mustar = out1[:,0]
    #         mustar = np.sort(mustar)
    #         import matplotlib.pyplot as plt
    #         plt.plot(mustar, 'ro')
    #         plt.show()

    # if False:
    #     # pass parameters and pool
    #     processes = 4
    #     pool = schwimmbad.choose_pool(mpi=False if csize==1 else True,
    #                                   processes=processes)
    #     func = jams.functions.morris
    #     npars = 20
    #     lb    = np.zeros(npars)
    #     ub    = np.ones(npars)
    #     beta0              = 0.
    #     beta1              = np.random.standard_normal(npars)
    #     beta1[:10]         = 20.
    #     beta2              = np.random.standard_normal((npars,npars))
    #     beta2[:6,:6]       = -15.
    #     beta3              = np.zeros((npars,npars,npars))
    #     beta3[:5,:5,:5]    = -10.
    #     beta4              = np.zeros((npars,npars,npars,npars))
    #     beta4[:4,:4,:4,:4] = 5.
    #     args = [beta0, beta1, beta2, beta3, beta4] # Morris
    #     nt = 20
    #     ntotal  = 10*nt
    #     nsteps = 6
    #     verbose = 1

    #     out = screening(func, lb, ub, nt, arg=args, mask=None,
    #                     nsteps=nsteps, ntotal=ntotal,
    #                     processes=processes, pool=pool, verbose=1)
    #     t2    = ptime.time()

    #     if crank == 0:
    #         strin = ('[m]: {:.1f}'.format(
    #                  (t2-t1)/60.)
    #                  if (t2-t1)>60. else '[s]: {:d}'.format(int(t2-t1)))
    #         print('Time elapsed: ', strin)
    #         print('mu*, mu, std: ', out)
    #         out1 = out / out.max(axis=0)
    #         mustar = out1[:,0]
    #         mustar = np.sort(mustar)
    #         import matplotlib.pyplot as plt
    #         plt.plot(mustar, 'ro')
    #         plt.show()
    #     pool.close()

    # # PYJAMS
    # from functools import partial
    # import numpy as np
    # from pyjams.functions import G, Gstar, K, fmorris
    # from partialwrap import function_wrapper

    # #
    # # G function
    # # seed for reproducible results
    # seed = 1234
    # np.random.seed(seed=seed)

    # func   = G
    # npars  = 6
    # params = [78., 12., 0.5, 2., 97., 33.] # G

    # # Partialise function with fixed parameters
    # arg   = [params]
    # kwarg = {}
    # obj = partial(function_wrapper, func, arg, kwarg)

    # # eee parameters
    # lb = np.zeros(npars)
    # ub = np.ones(npars)
    # nt      = 10
    # ntotal  = 50
    # nsteps  = 6
    # verbose = 1

    # out = ee(obj, lb, ub, nt, x0=None, mask=None, ntotal=ntotal,
    #          nsteps=nsteps, processes=4)
    # print('G')
    # print(np.around(out[:,0],3))

    # #
    # # G function
    # # seed for reproducible results
    # seed = 1234
    # np.random.seed(seed=seed)

    # func   = G
    # npars  = 6
    # params = [78., 12., 0.5, 2., 97., 33.] # G

    # # Partialise function with fixed parameters
    # arg   = [params]
    # kwarg = {}
    # obj = partial(function_wrapper, func, arg, kwarg)

    # # eee parameters
    # lb = np.zeros(npars)
    # ub = np.ones(npars)
    # nt      = 10
    # ntotal  = 50
    # nsteps  = 6
    # verbose = 1

    # out = ee(obj, lb, ub, nt, x0=None, mask=None, processes=4)
    # print('G')
    # print(np.around(out[:,0],3))

    # #
    # # Gstar function
    # # seed for reproducible results
    # seed = 1234
    # np.random.seed(seed=seed)

    # func   = Gstar
    # npars  = 10
    # params = [[np.ones(npars),     np.random.random(npars),
    #            [0., 0.,  9.,  9.,  9.,  9.,  9., 9., 9., 9.]],  # G*
    #           [np.ones(npars),     np.random.random(npars),
    #            [0., 0.1, 0.2, 0.3, 0.4, 0.8, 1., 2., 3., 4.]],
    #           [np.ones(npars)*0.5, np.random.random(npars),
    #            [0., 0.,  9.,  9.,  9.,  9.,  9., 9., 9., 9.]],
    #           [np.ones(npars)*0.5, np.random.random(npars),
    #            [0., 0.1, 0.2, 0.3, 0.4, 0.8, 1., 2., 3., 4.]],
    #           [np.ones(npars)*2.0, np.random.random(npars),
    #            [0., 0.,  9.,  9.,  9.,  9.,  9., 9., 9., 9.]],
    #           [np.ones(npars)*2.0, np.random.random(npars),
    #            [0., 0.1, 0.2, 0.3, 0.4, 0.8, 1., 2., 3., 4.]]]

    # # eee parameters
    # lb = np.zeros(npars)
    # ub = np.ones(npars)
    # nsteps = 6
    # verbose = 1

    # for ii in range(len(params)):
    #     # Partialise function with fixed parameters
    #     arg   = params[ii]
    #     kwarg = {}
    #     obj = partial(function_wrapper, func, arg, kwarg)

    #     out = ee(obj, lb, ub, nt, x0=None, mask=None, ntotal=ntotal,
    #              nsteps=nsteps, processes=4)
    #     print('G* ', ii)
    #     print(np.around(out[:,0],3))

    # #
    # # Bratley / K function
    # # seed for reproducible results
    # seed = 1234
    # np.random.seed(seed=seed)

    # func   = K
    # npars  = 10
    # params = [] # K

    # # eee parameters
    # lb = np.zeros(npars)
    # ub = np.ones(npars)
    # nsteps = 6
    # verbose = 1

    # out = ee(func, lb, ub, nt, x0=None, mask=None, ntotal=ntotal,
    #          nsteps=nsteps, processes=4)
    # print('K')
    # print(np.around(out[:,0],3))

    # #
    # # Morris function
    # # seed for reproducible results
    # seed = 1234
    # np.random.seed(seed=seed)

    # func = fmorris
    # npars = 20
    # beta0              = 0.
    # beta1              = np.random.standard_normal(npars)
    # beta1[:10]         = 20.
    # beta2              = np.random.standard_normal((npars,npars))
    # beta2[:6,:6]       = -15.
    # beta3              = np.zeros((npars,npars,npars))
    # beta3[:5,:5,:5]    = -10.
    # beta4              = np.zeros((npars,npars,npars,npars))
    # beta4[:4,:4,:4,:4] = 5.

    # # Partialise Morris function with fixed parameters beta0-4
    # arg   = [beta0, beta1, beta2, beta3, beta4]
    # kwarg = {}
    # obj = partial(function_wrapper, func, arg, kwarg)

    # # eee parameters
    # lb = np.zeros(npars)
    # ub = np.ones(npars)
    # nsteps = 6
    # verbose = 1

    # out = ee(obj, lb, ub, nt, x0=None, mask=None, ntotal=ntotal,
    #          nsteps=nsteps, processes=4)
    # print('Morris')
    # print(np.around(out[:,0],3))

    # from functools import partial
    # import numpy as np
    # from pyjams.functions import ishigami_homma
    # from partialwrap import function_mask_wrapper
    # seed = 1234
    # np.random.seed(seed=seed)

    # func   = ishigami_homma
    # npars  = 3

    # x0   = np.ones(npars)
    # mask = np.ones(npars, dtype=bool)
    # mask[1] = False

    # arg   = [1., 3.]
    # kwarg = {}

    # obj = partial(function_mask_wrapper, func, x0, mask, arg, kwarg)

    # lb = np.ones(npars) * (-np.pi)
    # ub = np.ones(npars) * np.pi
    # nt      = 10
    # ntotal  = 50
    # nsteps  = 6

    # out = ee(obj, lb[mask], ub[mask],
    #          nt, ntotal=ntotal, nsteps=nsteps,
    #          processes=1)

    # print('Ishigami-Homma')
    # print(np.around(out[:,0],3))

    # from functools import partial
    # import numpy as np
    # import scipy.stats as stats
    # from pyjams.functions import ishigami_homma
    # from partialwrap import function_mask_wrapper
    # seed = 1234
    # np.random.seed(seed=seed)

    # func   = ishigami_homma
    # npars  = 3

    # x0   = np.ones(npars)
    # mask = np.ones(npars, dtype=bool)
    # mask[1] = False

    # arg   = [1., 3.]
    # kwarg = {}

    # obj = partial(function_mask_wrapper, func, x0, mask, arg, kwarg)

    # lb = np.ones(npars) * (-np.pi)
    # ub = np.ones(npars) * np.pi
    # dist      = [ stats.uniform for i in range(npars) ]
    # distparam = [ (lb[i],ub[i]-lb[i]) for i in range(npars) ]
    # lb = np.zeros(npars)
    # ub = np.ones(npars)
    # nt      = 10
    # ntotal  = 50
    # nsteps  = 6

    # out = ee(obj, lb[mask], ub[mask],
    #          nt, ntotal=ntotal, nsteps=nsteps,
    #          dist=[ dist[i] for i in np.where(mask)[0] ],
    #          distparam=[ distparam[i] for i in np.where(mask)[0] ],
    #          processes=1)

    # print('Ishigami-Homma dist')
    # print(np.around(out[:,0],3))



#!/usr/bin/env python
"""
This is the unittest for color module.

python -m unittest -v tests/test_color.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_color.py

"""
from __future__ import division, absolute_import, print_function
import unittest


def _rgb2rgb(col):
    """
    Transform RGB tuple with values 0-255 to tuple with values 0-1
    """
    return tuple([ i / 255. for i in col ])


def _to_grey(col):
    """
    Transform RGB tuple to grey values
    """
    isgrey = 0.2125 * col[0] + 0.7154 * col[1] + 0.072 * col[2]
    return (isgrey, isgrey, isgrey)


class TestColor(unittest.TestCase):
    """
    Tests for color.py
    """

    def test_get_color(self):
        import matplotlib as mpl
        mpl.use('Agg')
        import matplotlib.pyplot as plt  # needed to work
        from pyjams.color import get_color
        from pyjams.color import ufz_colors

        # ufz
        ufz = ufz_colors

        # register ufz colors
        get_color()
        for cc in ufz:
            cmap = plt.cm.colors.get_named_colors_mapping()[cc]
            target = ufz[cc]
            self.assertEqual(cmap, target)

        # get single colors
        for cc in ufz:
            cmap = get_color(cc)
            target = ufz[cc]
            self.assertEqual(cmap, target)

        # get list of colors
        cmap = get_color(ufz.keys())
        target = list(ufz.values())
        self.assertEqual(cmap, target)

    def test_print_colors(self):
        from pyjams.color import print_colors

        assert print_colors('base') is None
        assert print_colors(['base', 'ufz']) is None
        assert print_colors() is None

    def test_get_cmap(self):
        import matplotlib as mpl
        mpl.use('Agg')
        import matplotlib.pyplot as plt  # needed to work
        from pyjams.color import get_cmap
        from pyjams.color import brewer_sequential
        from pyjams.color import mathematica_rainbow
        from pyjams.color import ncl_small
        from pyjams.color import pyjams_cmaps
        from pyjams.color import oregon_sequential
        from pyjams.color import sron2012_colors, sron2012_functions
        from pyjams.color import sron_colors, sron_colormaps, sron_functions

        # brewer
        brewer_ylgn3 = brewer_sequential['brewer_ylgn3']
        brewer_ylgn3 = [ _rgb2rgb(i) for i in brewer_ylgn3 ]

        # mathematica
        mathematica_dark_rainbow_8 = (
            mathematica_rainbow['mathematica_dark_rainbow_8'])

        # ncl
        ncl_amwg = ncl_small['ncl_amwg']

        # pyjams
        pyjams_amwg = pyjams_cmaps['pyjams_amwg']

        # oregon
        osu_bu7 = oregon_sequential['osu_bu7']
        osu_bu7 = [ _rgb2rgb(i) for i in osu_bu7 ]

        # sron2012
        sron2012_light = sron2012_colors['sron2012_light']
        sron2012_light = [ mpl.colors.colorConverter.to_rgb(i)
                           for i in sron2012_light ]

        sron2012_ylorbr_3 = [ sron2012_functions['sron2012_ylorbr'](i/2.)
                              for i in range(3) ]
        sron2012_buylrd_3 = [ sron2012_functions['sron2012_buylrd'](i/2.)
                              for i in range(3) ]
        sron2012_rainbow_3 = [ sron2012_functions['sron2012_rainbow'](i/2.)
                               for i in range(3) ]

        # sron
        sron_vibrant = sron_colors['sron_vibrant']
        sron_vibrant = [ mpl.colors.colorConverter.to_rgb(i)
                         for i in sron_vibrant ]

        sron_iridescent = sron_colormaps['sron_iridescent']
        sron_iridescent_miss = mpl.colors.colorConverter.to_rgba(
            sron_iridescent[1])
        sron_iridescent = [ mpl.colors.colorConverter.to_rgb(i)
                            for i in sron_iridescent[0] ]

        sron_rainbow_discrete = sron_functions['sron_rainbow_discrete']

        sron_rainbow_discrete_3 = sron_rainbow_discrete(3)
        sron_rainbow_discrete_3_miss = mpl.colors.colorConverter.to_rgba(
            sron_rainbow_discrete_3[1])
        sron_rainbow_discrete_3 = [ mpl.colors.colorConverter.to_rgb(i)
                                    for i in sron_rainbow_discrete_3[0] ]

        sron_rainbow_discrete_23 = sron_rainbow_discrete()
        sron_rainbow_discrete_23_miss = mpl.colors.colorConverter.to_rgba(
            sron_rainbow_discrete_23[1])
        sron_rainbow_discrete_23 = [ mpl.colors.colorConverter.to_rgb(i)
                                     for i in sron_rainbow_discrete_23[0] ]

        # brewer
        cmap = get_cmap('brewer_ylgn3')
        target = brewer_ylgn3
        self.assertEqual(cmap, target)

        # resample
        cmap = get_cmap('brewer_ylgn3', ncol=2)
        target = [brewer_ylgn3[0], brewer_ylgn3[2]]
        self.assertEqual(cmap, target)

        # color map
        cmap = get_cmap('brewer_ylgn3', as_cmap=True)
        assert isinstance(cmap, mpl.colors.ListedColormap)

        # reverse
        cmap = get_cmap('brewer_ylgn3', reverse=True)
        target = brewer_ylgn3[::-1]
        self.assertEqual(cmap, target)

        cmap = get_cmap('brewer_ylgn3_r')
        target = brewer_ylgn3[::-1]
        self.assertEqual(cmap, target)

        # grey
        cmap = get_cmap('brewer_ylgn3', grey=True)
        target = [ _to_grey(i) for i in brewer_ylgn3 ]
        self.assertEqual(cmap, target)

        # order
        cmap = get_cmap('brewer_ylgn3', order='HUE')
        target = brewer_ylgn3
        self.assertEqual(cmap, target)

        cmap = get_cmap('brewer_ylgn3', order='Saturation')
        target = brewer_ylgn3
        self.assertEqual(cmap, target)

        cmap = get_cmap('brewer_ylgn3', order='value')
        target = brewer_ylgn3[::-1]
        self.assertEqual(cmap, target)

        cmap = get_cmap('brewer_ylgn3', order='value', reverse=True)
        target = brewer_ylgn3
        self.assertEqual(cmap, target)

        # mathematica
        cmap = get_cmap('mathematica_dark_rainbow_8')
        target = mathematica_dark_rainbow_8
        self.assertEqual(cmap, target)

        # ncl
        cmap = get_cmap('ncl_amwg')
        target = ncl_amwg
        self.assertEqual(cmap, target)

        # pyjams
        cmap = get_cmap('pyjams_amwg')
        target = pyjams_amwg
        self.assertEqual(cmap, target)

        # oregon
        cmap = get_cmap('osu_bu7')
        target = osu_bu7
        self.assertEqual(cmap, target)

        # sron2012_colors
        cmap = get_cmap('sron2012_light')
        target = sron2012_light
        self.assertEqual(cmap, target)

        # sron2012_functions
        cmap = get_cmap('sron2012_ylorbr', ncol=3)
        target = sron2012_ylorbr_3
        self.assertEqual(cmap, target)

        # sron2012_functions
        cmap = get_cmap('sron2012_buylrd', ncol=3)
        target = sron2012_buylrd_3
        self.assertEqual(cmap, target)

        # sron2012_functions
        cmap = get_cmap('sron2012_rainbow', 3)
        target = sron2012_rainbow_3
        self.assertEqual(cmap, target)

        # sron2012_functions - ncol=0
        cmap = get_cmap('sron2012_rainbow')
        target = [ sron2012_functions['sron2012_rainbow'](i/255.)
                   for i in range(256) ]
        self.assertEqual(cmap, target)

        # sron2012_functions - offset, upper
        cmap = get_cmap('sron2012_rainbow', ncol=3, offset=0.2, upper=0.9)
        target = [ sron2012_functions['sron2012_rainbow'](0.2+i*0.7/2.)
                   for i in range(3) ]
        self.assertEqual(cmap, target)

        # sron_colors
        cmap = get_cmap('sron_vibrant')
        target = sron_vibrant
        self.assertEqual(cmap, target)

        # sron_colormaps
        cmap = get_cmap('sron_iridescent')
        target = sron_iridescent
        self.assertEqual(cmap, target)

        # sron_colormaps - as_cmap, miss
        cmap = get_cmap('sron_iridescent', as_cmap=True)
        cols = cmap.colors
        target = sron_iridescent
        self.assertEqual(cols, target)
        if mpl.__version__ > '3.4.0':
            miss = tuple(cmap.get_bad())
            target = sron_iridescent_miss
            self.assertEqual(miss, target)

        # sron_functions
        cmap = get_cmap('sron_rainbow_discrete', 3)
        target = sron_rainbow_discrete_3
        self.assertEqual(cmap, target)
        cmap = get_cmap('sron_rainbow_discrete', 23)
        target = sron_rainbow_discrete_23
        self.assertEqual(cmap, target)

        # sron_functions - as_cmap, miss
        cmap = get_cmap('sron_rainbow_discrete', ncol=3, as_cmap=True)
        cols = cmap.colors
        target = sron_rainbow_discrete_3
        self.assertEqual(cols, target)
        if mpl.__version__ > '3.4.0':
            miss = tuple(cmap.get_bad())
            target = sron_rainbow_discrete_3_miss
            self.assertEqual(miss, target)
        cmap = get_cmap('sron_rainbow_discrete', as_cmap=True)
        cols = cmap.colors
        target = sron_rainbow_discrete_23
        self.assertEqual(cols, target)
        if mpl.__version__ > '3.4.0':
            miss = tuple(cmap.get_bad())
            target = sron_rainbow_discrete_23_miss
            self.assertEqual(miss, target)

        # matplotlib - ListedColormap
        cmap = get_cmap('viridis')
        target = mpl.cm.get_cmap('viridis').colors
        self.assertEqual(cmap, target)

        # matplotlib - LinearSegmentedColormap
        cmap = get_cmap('Blues')
        target = mpl.cm.get_cmap('Blues')
        target = [ target(i) for i in range(target.N) ]
        self.assertEqual(cmap, target)

        # matplotlib - ListedColormap - upper-/lowercase
        cmap = get_cmap('Viridis')
        target = mpl.cm.get_cmap('viridis').colors
        self.assertEqual(cmap, target)

        # matplotlib - LinearSegmentedColormap - upper-/lowercase
        cmap = get_cmap('blues')
        target = mpl.cm.get_cmap('Blues')
        target = [ target(i) for i in range(target.N) ]
        self.assertEqual(cmap, target)

        # nonsense name
        self.assertRaises(ValueError, get_cmap, 'NonSense')

        # unknown order
        self.assertRaises(ValueError, get_cmap, 'brewer_ylgn3',
                          order='luminance')

    def test_print_palettes(self):
        from pyjams.color import print_palettes

        assert print_palettes('mathematica') is None
        assert print_palettes(['mathematica', 'matplotlib']) is None
        assert print_palettes() is None


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Saturation vapour pressure over water and ice

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   esat

History
    * Written Jan 2012 by Matthias Cuntz (mc (at) macu (dot) de)
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Changed handling of masked arrays, Oct 2013, Matthias Cuntz
    * Assert T>0, Apr 2014, Matthias Cuntz
    * Using numpy docstring format, May 2020, Matthias Cuntz
    * Ported to pyjams, Jan 2022, Matthias Cuntz
    * Return same type as input type, Jan 2022, Matthias Cuntz
    * Raise AssertionError rather than assuming Celcius,
      Jan 2022, Matthias Cuntz
    * Use warnings for T < 100 K, Jan 2022, Matthias Cuntz
    * Change handling of return type to allow more (unspecific) iterable types
      such as pandas time series, Jan 2022, Matthias Cuntz
    * Return numpy array if type(input)(output) fails for unknown iterable
      types, Jan 2022, Matthias Cuntz
    * Use helper functions input2array and array2input,
      Jan 2022, Matthias Cuntz

"""
from collections.abc import Iterable
import numpy as np
from .helper import input2array, array2input


__all__ = ['esat']


def esat(T, formula='GoffGratch', undef=-9999., liquid=False):
    """
    Calculates the saturation vapour pressure over water and/or ice

    For temperatures above (and equal to) 0 degree C (273.15 K),
    the vapour pressure over liquid water is calculated;
    for temperatures below 0 degree C, the vapour pressure over ice
    is calculated.

    Parameters
    ----------
    T : float or array_like
        Temperature [K]
    formula : str, optional
        Formula name to use for calculations; case-insensitive.
        Available formulations are:
        GoffGratch (default) [1]_, WMO [2]_, IAPWS [3]_, MagnusTeten [4]_,
        MartiMauersberger [5]_, Buck_original [6]_, Buck [7]_, Wexler [8]_,
        Sonntag [9]_, Bolton [10]_, Fukuta [11]_, HylandWexler [12]_,
        MurphyKoop [13]_

        Note that several formulations do not provide a vapour pressure
        formulation over ice and Marti and Mauersberger do not provide
        a formula over liquid: GoffGratch is used in theses cases.
    undef : float, optional
        Exclude `T == undef` from calculations (default: -9999.)
    liquid : bool, optional
        If True, use liquid formula for all temperatures.

    Returns
    -------
    float or array_like
        Saturation water pressure at temperature T in Pascal [Pa].

    Notes
    -----
    Taken from Holger Voemel:
    http://cires.colorado.edu/~voemel/vp.html

    References
    ----------
    .. [1] Goff Gratch formulation
       Smithsonian Meteorological Tables, 5th edition, p. 350, 1984
       Original source: Goff and Gratch (1946), p. 107

    .. [2] Intended WMO formulation, originally published by Goff (1957)
       incorrectly referenced by WMO technical regulations,
       WMO-NO 49, Vol I, General Meteorological Standards and
       Recommended Practices, App. A, 1988, Corrigendum Aug 2000

    .. [3] Wagner W. and A. Pruss (2002)
       The IAPWS formulation 1995 for the thermodynamic properties
       of ordinary water substance for general and scientific use,
       J. Phys. Chem. Ref. Data, 31(2), 387-535
       This is the 'official' formulation from the International
       Association for the Properties of Water and Steam
       The valid range of this formulation is 273.16 <= T <= 647.096 K
       and is based on the ITS90 temperature scale.

    .. [4] Murray, F. W., On the computation of saturation vapour pressure,
       J. Appl. Meteorol., 6, 203-204, 1967

    .. [5] Marti, J. and K Mauersberger, A survey and new measurements of
       ice vapour pressure at temperatures between 170 and 250 K,
       GRL 20, 363-366, 1993

    .. [6] Bucks vapour pressure formulation based on Tetens formula
       Buck, A. L., New equations for computing vapour pressure and
       enhancement factor, J. Appl. Meteorol., 20, 1527-1532, 1981

    .. [7] Bucks vapour pressure formulation based on Tetens formula
       Buck Research, Model CR-1A Hygrometer Operating Manual, Sep 2001

    .. [8] Wexler, A., Vapour pressure formulation for ice
       Journal of Research of the National Bureau of Standards-A.
       81A, 5-20, 1977

    .. [9] Sonntag, D., Advancements in the field of hygrometry,
       Meteorol. Z., N. F., 3, 51-66, 1994

    .. [10] Bolton, D., The computation of equivalent potential temperature
       Monthly Weather Report, 108, 1046-1053, 1980. equation (10)

    .. [11] Fukuta, N. and C. M. Gramada, Vapour pressure measurement of
       supercooled water, J. Atmos. Sci., 60, 1871-1875, 2003
       This paper does not give a vapour pressure formulation,
       but rather a correction over the Smithsonian Tables.
       Thus calculate the table value first, then use the correction
       to get to the measured value.
       This is done only for -39 < TC < 0.

    .. [12] Hyland, R. W. and A. Wexler, Formulations for the Thermodynamic
       Properties of the saturated Phases of H2O from 173.15K to
       473.15K, ASHRAE Trans, 89(2A), 500-519, 1983

    .. [13] Murphy and Koop, Review of the vapour pressure of ice and
       supercooled water for atmospheric applications
       Q. J. R. Meteorol. Soc (2005), 131, pp. 1539-1565

    Examples
    --------
    >>> print('{:.3f}'.format(esat(293.15)))
    2335.847
    >>> print('{:.3f}'.format(esat(253.15)))
    103.074

    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15])))
    2335.847 103.074
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],
    ...     formula='GoffGratch')))
    2335.847 103.074
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],
    ...     formula='MartiMauersberger')))
    2335.847 103.650
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],
    ...     formula='MagnusTeten')))
    2335.201 102.771
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],formula='buck')))
    2338.340 103.286
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],
    ...     formula='Buck_original')))
    2337.282 103.267
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],formula='wmo')))
    2337.080 103.153
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],formula='WEXLER')))
    2323.254 103.074
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],formula='Sonntag')))
    2339.249 103.249
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],formula='Bolton')))
    2336.947 103.074
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],formula='Fukuta')))
    2335.847 103.074
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],
    ...     formula='HylandWexler')))
    2338.804 103.260
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],formula='IAPWS')))
    2339.194 103.074
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15],
    ...     formula='MurphyKoop')))
    2339.399 103.252

    >>> print('{:.3f} {:.3f}'.format(*esat(np.array([293.15,253.15]),
    ...     liquid=True)))
    2335.847 125.292
    >>> print('{:.3f} {:.3f}'.format(*esat([293.15,253.15], formula='Fukuta',
    ...     liquid=True)))
    2335.847 125.079

    >>> out = esat(np.ma.array([253.15,-9999.], mask=[False,True]))
    >>> print('{:.3f} {:.3f}'.format(*out.filled(-9999.)))
    103.074 -9999.000

    """
    # Constants
    T0 = 273.15  # Celcius <-> Kelvin [K]
    knownforms = ['Buck', 'Buck_original', 'Bolton', 'Fukuta', 'GoffGratch',
                  'HylandWexler', 'IAPWS', 'MagnusTeten', 'MartiMauersberger',
                  'MurphyKoop', 'Sonntag', 'Vaisala', 'Wexler', 'WMO']
    lknown = [ i.lower() for i in knownforms ]

    # Check input type
    mT = input2array(T, undef=undef, default=T0)
    # Check unfeasible temperatures
    assert np.ma.all(mT > 0.), (
        'Temperature below 0 K; probably given in Celsius instead of Kelvin.')
    assert np.ma.all(mT < T0 + 100.), ('Temperature above 373.15 K')
    if np.ma.any(mT < 100.):
        import warnings
        warnings.warn('Temperature < 100 K; probably given in Celsius'
                      ' instead of Kelvin.', UserWarning)
    # Check formulas
    form = formula.lower()
    if form not in lknown:
        raise ValueError('Formula not known. Known formulas are'
                         ' (default: GoffGratch):'
                         '{:s}'.format(', '.join(knownforms)))

    # Split input into masked arrays
    Tlim = 1e-3 if liquid else T0
    if mT.size > 1:
        ii = np.ma.where(mT >= Tlim)[0]
        jj = np.ma.where(mT < Tlim)[0]
        if ii.size > 0:
            T_liq  = mT[ii]
        if jj.size > 0:
            T_ice  = mT[jj]
    else:
        if mT >= Tlim:
            ii = np.array([0])
            jj = np.array([])
            T_liq = mT
        else:
            ii = np.array([])
            jj = np.array([0])
            T_ice = mT
    out = mT.copy()  # to conserve mask

    #
    # Calc
    #

    # Liquid
    if ii.size > 0:
        TC_liq = T_liq - T0
        if form == 'buck':
            esat_liq = 6.1121 * np.ma.exp(
                (18.678 - (TC_liq) / 234.5) * (TC_liq) / (257.14+TC_liq))
        elif form == 'buck_original':
            esat_liq = 6.1121 * np.ma.exp(17.502 * TC_liq / (240.97+TC_liq))
        elif form == 'bolton':
            esat_liq = 6.112 * np.ma.exp(17.67 * TC_liq / (TC_liq+243.5))
        elif form == 'fukuta':
            # steam point temperature in K
            Ts  = 373.16
            # saturation pressure at steam point temperature, normal atmosphere
            ews = 1013.246
            esat_liq = (10.**(-7.90298 * (Ts/T_liq-1.)
                              + 5.02808 * np.ma.log10(Ts/T_liq)
                              - 1.3816e-7 * (10.**(11.344*(1.-T_liq/Ts))-1.)
                              + 8.1328e-3 * (10.**(-3.49149*(Ts/T_liq-1))-1.)
                              + np.ma.log10(ews)))
            mm = (TC_liq < 0.) & (TC_liq > -39.)
            if np.ma.any(mm):
                x = TC_liq + 19.
                esat_liq = (np.where(mm,
                                     esat_liq * (0.9992
                                                 + 7.113e-4*x
                                                 - 1.847e-4*x**2
                                                 + 1.189e-5*x**3
                                                 + 1.130e-7*x**4
                                                 - 1.743e-8*x**5),
                                     esat_liq))
        elif ((form == 'goffgratch') | (form == 'martimauersberger')):
            # steam point temperature in K
            Ts  = 373.16
            # saturation pressure at steam point temperature, normal atmosphere
            ews = 1013.246
            esat_liq = (10.**(-7.90298*(Ts/T_liq-1.)
                              + 5.02808 * np.ma.log10(Ts/T_liq)
                              - 1.3816e-7 * (10.**(11.344*(1.-T_liq/Ts))-1.)
                              + 8.1328e-3 * (10.**(-3.49149*(Ts/T_liq-1.))-1.)
                              + np.ma.log10(ews)))
        elif form == 'hylandwexler':
            esat_liq = (np.ma.exp(- 0.58002206e4/T_liq
                                  + 0.13914993e1 - 0.48640239e-1*T_liq
                                  + 0.41764768e-4*T_liq**2
                                  - 0.14452093e-7*T_liq**3
                                  + 0.65459673e1 * np.ma.log(T_liq)) / 100.)
        elif form == 'iapws':
            Tc = 647.096      # K   : Temperature at the critical point
            Pc = 22.064e4     # hPa : Vapour pressure at the critical point
            nu = (1. - T_liq/Tc)
            a1 = -7.85951783
            a2 = 1.84408259
            a3 = -11.7866497
            a4 = 22.6807411
            a5 = -15.9618719
            a6 = 1.80122502
            esat_liq = Pc * np.ma.exp(Tc/T_liq
                                      * (a1*nu + a2*nu**1.5
                                         + a3*nu**3 + a4*nu**3.5
                                         + a5*nu**4 + a6*nu**7.5))
        elif form == 'magnusteten':
            esat_liq = 10.**(7.5 * (TC_liq) / (TC_liq+237.5) + 0.7858)
        elif form == 'murphykoop':
            esat_liq = (np.exp(54.842763 - 6763.22 / T_liq
                               - 4.210 * np.ma.log(T_liq)
                               + 0.000367 * T_liq
                               + np.tanh(0.0415 * (T_liq - 218.8))
                               * (53.878 - 1331.22 / T_liq - 9.44523
                                  * np.ma.log(T_liq) + 0.014025 * T_liq))
                        / 100.)
        elif form == 'sonntag':
            esat_liq = (np.ma.exp(- 6096.9385 * 1./T_liq
                                  + 16.635794
                                  - 2.711193e-2 * T_liq
                                  + 1.673952e-5 * T_liq**2
                                  + 2.433502 * np.ma.log(T_liq)))
        elif form == 'wexler':
            esat_liq = (np.ma.exp(- 2.9912729e3 * 1./T_liq**2
                                  - 6.0170128e3 * 1./T_liq
                                  + 1.887643854e1
                                  - 2.8354721e-2 * T_liq**1
                                  + 1.7838301e-5 * T_liq**2
                                  - 8.4150417e-10 * T_liq**3
                                  - 4.4412543e-13 * T_liq**4
                                  + 2.858487 * np.ma.log(T_liq)) / 100.)
        elif form == 'wmo':
            Ts   = 273.16  # steam point temperature in K
            esat_liq = (10.**(10.79574 * (1.-Ts/T_liq)
                              - 5.02800 * np.ma.log10(T_liq/Ts)
                              + 1.50475e-4 * (1.-10.**(-8.2969*(T_liq/Ts-1.)))
                              + 0.42873e-3 * (10.**(+4.76955*(1.-Ts/T_liq))-1.)
                              + 0.78614))
        else:
            # should not come here, hence no cover
            raise ValueError('Formula not known for liquid:'
                             ' {:s}'.format(formula))  # pragma: no cover
        esat_liq *= 100.  # hPa -> Pa
        if isinstance(T, Iterable):
            out[ii] = esat_liq
        else:
            out = esat_liq

    # Ice
    if jj.size > 0:
        TC_ice = T_ice - T0
        if form == 'buck':
            esat_ice = 6.1115 * np.exp((23.036 - TC_ice / 333.7)
                                       * TC_ice / (279.82+TC_ice))
        elif form == 'buck_original':
            esat_ice = 6.1115 * np.exp(22.452 * TC_ice / (272.55+TC_ice))
        elif ((form == 'goffgratch') | (form == 'bolton')
              | (form == 'fukuta') | (form == 'iapws') | (form == 'wexler')):
            ei0 = 6.1071  # mbar
            Ts  = 273.16  # freezing point in K
            esat_ice = np.ma.exp(np.log(10.)*(-9.09718 * (Ts/T_ice-1.)
                                              - 3.56654 * np.ma.log10(Ts/T_ice)
                                              + 0.876793 * (1.-T_ice/Ts)
                                              + np.log10(ei0)))
        elif (form == 'hylandwexler'):
            esat_ice = (np.exp(- 0.56745359E4 / T_ice
                               + 0.63925247E1
                               - 0.96778430E-2 * T_ice
                               + 0.62215701E-6 * T_ice**2
                               + 0.20747825E-8 * T_ice**3
                               - 0.94840240E-12 * T_ice**4
                               + 0.41635019E1 * np.log(T_ice)) / 100.)
        elif form == 'magnusteten':
            esat_ice = 10.**(9.5 * TC_ice/(265.5+TC_ice) + 0.7858)
        elif form == 'martimauersberger':
            esat_ice = 10.**(-2663.5/T_ice + 12.537) / 100.
        elif form == 'murphykoop':
            esat_ice = np.exp(9.550426 - 5723.265/T_ice
                              + 3.53068 * np.log(T_ice)
                              - 0.00728332 * T_ice) / 100.
        elif form == 'sonntag':
            esat_ice = (np.exp(- 6024.5282 * 1./T_ice
                               + 24.721994
                               + 1.0613868E-2 * T_ice
                               - 1.3198825E-5 * T_ice**2
                               - 0.49382577 * np.log(T_ice)))
        elif form == 'wmo':
            Ts = 273.16  # steam point temperature in K
            esat_ice = (10.**(-9.09685 * (Ts/T_ice-1.)
                              - 3.56654 * np.log10(Ts/T_ice)
                              + 0.87682 * (1.-T_ice/Ts) + 0.78614))
        else:
            # should not come here, hence no cover
            raise ValueError('Formula not known for ice:'
                             ' {:s}'.format(formula))  # pragma: no cover
        esat_ice *= 100.  # hPa -> Pa
        if isinstance(T, Iterable):
            out[jj] = esat_ice
        else:
            out = esat_ice

    # return same type as input type
    out = array2input(out, T, undef=undef)

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
Helper functions for pyjams library

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2022- Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   array2input
   input2array

History
    * Written input2array and array2input, Jan 2022, Matthias Cuntz
    * Added isundef, Mar 2022, Matthias Cuntz
    * Second input to array2input, Mar 2022, Matthias Cuntz
    * undef=np.nan by default, Apr 2022, Matthias Cuntz
    * Array masked or set to undef only if shapes of array and input agree in
      array2input, Apr 2022, Matthias Cuntz
    * Allow string arrays, Jun 2022, Matthias Cuntz
    * Allow undef=='' in isundef, Jun 2022, Matthias Cuntz
    * Single string input as 1-element array, Jun 2022, Matthias Cuntz
    * Allow undef=None, Jun 2022, Matthias Cuntz
    * undef=None by default, Jun 2022, Matthias Cuntz
    * Refine using undef=None with numpy.array and list,
      Jun 2022, Matthias Cuntz

"""
from collections.abc import Iterable
import numpy as np


__all__ = ['isundef', 'array2input', 'input2array']


def isundef(arr, undef):
    """
    Check if *arr* is undef

    Return arr==undef, taking care of NaN and Inf.

    Parameters
    ----------
    arr : scalar or numpy array
        Input scalar array
    undef : object
        Check if *arr == undef*.
        It is also possible to use None, '', np.nan, or np.inf for *undef*.

    Returns
    -------
    numpy array with *arr==undef*

    Examples
    --------
    >>> inp = [253.15, -9999.]
    >>> print(isundef(inp, -9999.))
    [False True]

    """
    if undef is None:
        return False
    elif not undef:
        return arr == undef
    elif np.isnan(undef):
        return np.isnan(arr)
    elif np.isinf(undef):
        return np.isinf(arr)
    else:
        return arr == undef


def array2input(outin, inp, inp2=None, undef=None):
    """
    Transforms numpy array to same type as input

    The numpy array *outin* will be transformed to the same type as *inp*.
    If shapes agree then masked values on *inp* will be masked on output
    and undefined values in *inp* will result in undefined output values.

    If *inp2* is given, then type of *inp* will take precedence,
    except if *inp* is a scalar or *inp2* is a masked array in which case
    the type of *inp2* will be taken.

    The function is supposed to work with :func:`input2array`, which makes the
    input a numpy array, setting masked values and undefined values to some
    default values to avoid math over- and underflow. `array2input`
    transforms the output back to the input format.

    Parameters
    ----------
    outin : numpy array
        Input array
    inp : scalar or iterable of numbers
        Original input variable that was transformed to numpy array with
        :func:`input2array`
    inp2 : scalar or iterable of numbers, optional
        Second input variable that was transformed to numpy array with
        :func:`input2array` (default: None)
    undef : float, optional
        Values in *inp* having value *undef* will result in ouput set to
        *undef* (default: None)

    Returns
    -------
    *outin* as same type as *inp*, or the type of *inp2* if *inp* is
    a scalar or *inp2* is a masked array

    Examples
    --------
    >>> inp = [253.15, -9999.]
    >>> inarray = input2array(inp, undef=-9999., default=273.15)
    >>> print(array2input(inarray, inp, undef=-9999.))
    [253.15 -9999.]

    """
    if inp2 is not None:
        if ( (not isinstance(inp, Iterable)) or
             isinstance(inp2, np.ma.MaskedArray) ):
            return array2input(outin, inp2, undef=undef)

    if isinstance(inp, Iterable):
        if isinstance(inp, np.ma.MaskedArray):
            if np.array(outin).shape == inp.shape:
                outout = np.ma.array(outin,
                                     mask=(isundef(inp, undef) | (inp.mask)))
            else:
                if isinstance(outin, np.ma.MaskedArray):
                    outout = outin
                else:
                    outout = np.ma.array(outin)
        elif isinstance(inp, np.ndarray):
            if np.array(outin).shape == inp.shape:
                if np.any(isundef(inp, undef)):
                    outout = np.where(isundef(inp, undef), undef, outin)
                else:
                    if isinstance(outin, np.ndarray):
                        outout = outin
                    else:
                        outout = np.array(outin)
            else:
                if isinstance(outin, np.ndarray):
                    outout = outin
                else:
                    outout = np.array(outin)
        elif isinstance(inp, str):
            if isundef(inp, undef):
                outout = undef
            else:
                if isinstance(outin, str):
                    outout = outin
                else:
                    outout = outin[0]
        else:
            if np.array(outin).shape == np.array(inp).shape:
                if np.any(isundef(np.array(inp), undef)):
                    outout = np.where(isundef(np.array(inp), undef),
                                      undef, outin)
                else:
                    outout = outin
            else:
                outout = outin
            try:
                outout = type(inp)(outout)
            except:  # pragma: no cover
                # unknown iterables so no cover
                pass
    else:
        if isundef(inp, undef):
            outout = undef
        else:
            try:
                outout = type(inp)(outin)
            except:  # pragma: no cover
                # unknown iterables so no cover
                outout = outin

    return outout


def input2array(inp, undef=None, default=1):
    """
    Makes numpy array from iterable or scalar input with masked or undef values
    are set to a default value

    The input variable will be transformed to a numpy array so that numpy
    functions and similar will work on all input. Undefined and masked values
    will be set to a default value to avoid math over- and underflow.

    The function is supposed to work with :func:`array2input`, which sets
    the output to the same type as the input; masked values on input will be
    masked on output, undefined input values will result in undefined output
    values.

    Parameters
    ----------
    inp : scalar or iterable of numbers
        Input variable to transform to numpy array
    undef : float, optional
        Values in *inp* having value *undef* will be set to *default*
        (default: None)
    default : number
        Values in *inp* having value *undef* will be set to *default*
        (default: 1)

    Returns
    -------
    ndarray
        Input variable transformed to numpy array with values *undef* set to
        *default*

    Examples
    --------
    >>> print(input2array([253.15, -9999.], undef=-9999., default=273.15))
    [253.15 273.15]

    """
    if isinstance(inp, Iterable):
        if isinstance(inp, np.ma.MaskedArray):
            out = np.ma.where(isundef(inp, undef), default,
                              inp).filled(default)
        elif isinstance(inp, str):
            out = np.array([inp])
            out = np.where(isundef(out, undef), default, out)
        else:
            out = np.array(inp)
            out = np.where(isundef(out, undef), default, out)
    else:
        # scalar
        out = np.array(default) if isundef(inp, undef) else np.array(inp)

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
This is the unittest for alpha_equ_h2o module.

python -m unittest -v tests/test_alpha_equ_h2o.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_alpha_equ_h2o.py

"""
import unittest


class TestAlphaEquH2O(unittest.TestCase):
    """
    Tests for alpha_equ_h2o.py
    """

    def test_alpha_equ_h2o(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import alpha_equ_h2o

        T0 = 273.15

        # scalar
        assert isinstance(alpha_equ_h2o(T0), float)
        assert np.around(alpha_equ_h2o(T0, isotope=1), 4) == 1.1123

        # list
        T = [0., 10., 15., 25.]
        T1 = [ tt + T0 for tt in T ]
        assert isinstance(alpha_equ_h2o(T1), list)
        self.assertEqual(list(alpha_equ_h2o(T1, isotope=0)),
                         [1.0, 1.0, 1.0, 1.0])

        # tuple
        T1 = tuple(T1)
        assert isinstance(alpha_equ_h2o(T1), tuple)
        self.assertEqual(list(np.around(alpha_equ_h2o(T1, isotope=1), 4)),
                         [1.1123, 1.0977, 1.0911, 1.0793])

        # ndarray
        T  = np.array(T)
        assert isinstance(alpha_equ_h2o(T+T0), np.ndarray)
        self.assertEqual(list(alpha_equ_h2o(T+T0, isotope=0)),
                         [1.0, 1.0, 1.0, 1.0])
        self.assertEqual(list(np.around(alpha_equ_h2o(T+T0, isotope=2), 4)),
                         [1.0117, 1.0107, 1.0102, 1.0094])

        # masked_array and greater1=False
        alpha = alpha_equ_h2o(np.ma.array(T+T0, mask=(T == 0.)),
                              isotope=2, greater1=False)
        assert isinstance(alpha, np.ndarray)
        assert isinstance(alpha, np.ma.MaskedArray)
        self.assertEqual(list(np.around(alpha, 4)),
                         [masked, 0.9894, 0.9899, 0.9907])

        # epsilon, HO18O, scalar
        epsilon = alpha_equ_h2o(0.+T0, isotope=2, eps=True) * 1000.
        self.assertEqual(np.around(epsilon, 4), 11.7187)

        # epsilon, HDO
        epsilon = alpha_equ_h2o(T+T0, isotope=1, eps=True) * 1000.
        self.assertEqual(list(np.around(epsilon, 4)),
                         [112.3194, 97.6829, 91.1296, 79.3443])

        # undef, scalar
        epsilon = alpha_equ_h2o(-9998., undef=-9998., isotope=2, eps=True)
        self.assertEqual(epsilon, -9998.)
        epsilon = alpha_equ_h2o(-9999., isotope=2, eps=True)
        self.assertEqual(epsilon, -9999.)

        # undef is NaN, scalar
        epsilon = alpha_equ_h2o(np.nan, undef=np.nan, isotope=2, eps=True)
        assert np.isnan(epsilon)
        epsilon = alpha_equ_h2o(np.nan, isotope=2, eps=True)
        assert np.isnan(epsilon)

        # undef is Inf, scalar
        epsilon = alpha_equ_h2o(np.inf, undef=np.inf, isotope=2, eps=True)
        assert np.isinf(epsilon)

        # undef, list
        T1 = list(T1)
        T1[0] = -1.
        alpha = alpha_equ_h2o(T1, undef=-1., isotope=1, eps=True)
        alpha = [ a * 1000. for a in alpha ]
        self.assertEqual(list(np.around(alpha, 4)),
                         [-1000.0000, 97.6829, 91.1296, 79.3443])

        # undef, tuple
        T1 = tuple(T1)
        alpha = alpha_equ_h2o(T1, undef=-1., isotope=1, eps=True)
        alpha = [ a * 1000. for a in alpha ]
        self.assertEqual(list(np.around(alpha, 4)),
                         [-1000.0000, 97.6829, 91.1296, 79.3443])

        # undef, np.array
        T1 = np.array(T1)
        alpha = alpha_equ_h2o(T1, undef=-1., isotope=1, eps=True) * 1000.
        self.assertEqual(list(np.around(alpha, 4)),
                         [-1000.0000, 97.6829, 91.1296, 79.3443])

        # undef, masked_array
        T1 = np.ma.array(T1, mask=(T == 10.))
        alpha = alpha_equ_h2o(T1, undef=-1., isotope=1, eps=True) * 1000.
        self.assertEqual(list(np.around(alpha, 4)),
                         [masked, masked, 91.1296, 79.3443])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Conversion of datetime formats

The module enhances cftime by non-CF date formats.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2022- Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   date2dec
   date2num
   dec2date
   num2date
   datetime

History
    * Written Jun 2022, Matthias Cuntz

ToDo
    * make PR cftime - add to_tuple() to real_datetime
    * make PR cftime - missing precision in _idealized_calendars
    * Add old history from JAMS
    * Complete test_datetime.py
    * add date2index
    * add time2index
    * implement fromordinal
    * implement change_calendar

"""
import re
import time as ptime
from datetime import datetime as datetime_python
from datetime import timedelta
import numpy as np
import cftime as cf
from .helper import input2array, array2input
from .date2date import date2date
# from pyjams.helper import input2array, array2input
# from pyjams.date2date import date2date


# from ._cftime import num2date, date2num, date2index, time2index, num2pydate
__all__ = ['date2dec', 'date2num', 'date2dec', 'num2date', 'datetime']


# supported calendars. Includes synonyms ('excel'=='excel1900')
_excelcalendars = ['excel', 'excel1900', 'excel1904']
_decimalcalendars = ['decimal', 'decimal360', 'decimal365', 'decimal366']
_noncfcalendars = _excelcalendars + _decimalcalendars
_cfcalendars = ['standard', 'gregorian', 'proleptic_gregorian',
                'noleap', 'julian', 'all_leap', '365_day', '366_day',
                '360_day']
_idealized_cfcalendars = ['all_leap', 'noleap', '366_day', '365_day',
                          '360_day']

# number of days in year
_dayspermonth      = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
_dayspermonth_leap = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
_dayspermonth_360  = [30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30]
_cumdayspermonth      = [0, 31, 59, 90, 120, 151, 181,
                         212, 243, 273, 304, 334, 365]
_cumdayspermonth_leap = [0, 31, 60, 91, 121, 152, 182,
                         213, 244, 274, 305, 335, 366]
_cumdayspermonth_360  = [0, 30, 60, 90, 120, 150, 180,
                         210, 240, 270, 300, 330, 360]

#
# Exact copies of private cftime routines
# (changed Python time to ptime)
#

_illegal_s = re.compile(r"((^|[^%])(%%)*%s)")


def _findall(text, substr):
    # Also finds overlaps
    sites = []
    i = 0
    while 1:
        j = text.find(substr, i)
        if j == -1:
            break
        sites.append(j)
        i = j + 1
    return sites


# Every 28 years the calendar repeats, except through century leap
# years where it's 6 years.  But only if you're using the Gregorian
# calendar.  ;)
def _strftime(dt, fmt):
    if _illegal_s.search(fmt):
        raise TypeError("This strftime implementation does not handle %s")
    # don't use strftime method at all.
    # if dt.year > 1900:
    #    return dt.strftime(fmt)

    year = dt.year
    # For every non-leap year century, advance by
    # 6 years to get into the 28-year repeat cycle
    delta = 2000 - year
    off = 6 * (delta // 100 + delta // 400)
    year = year + off

    # Move to around the year 2000
    year = year + ((2000 - year) // 28) * 28
    timetuple = dt.timetuple()
    s1 = ptime.strftime(fmt, (year,) + timetuple[1:])
    sites1 = _findall(s1, str(year))

    s2 = ptime.strftime(fmt, (year + 28,) + timetuple[1:])
    sites2 = _findall(s2, str(year + 28))

    sites = []
    for site in sites1:
        if site in sites2:
            sites.append(site)

    s = s1
    syear = "%04d" % (dt.year,)
    for site in sites:
        s = s[:site] + syear + s[site + 4:]
    return s


# factory function without optional kwargs that can be used in
# datetime.__reduce_
def _create_datetime(date_type, args, kwargs):
    return date_type(*args, **kwargs)


#
# Adapted cftime routines
#

def _datesplit(timestr):
    """
    Split a unit string for time into its three components:
    unit, string 'since' or 'as', and the remainder

    """
    try:
        (units, sincestring, remainder) = timestr.split(None, 2)
    except ValueError:
        raise ValueError(f'Incorrectly formatted date-time unit_string:'
                         f' {timestr}')

    if sincestring.lower() not in ['since', 'as']:
        raise ValueError(f"No 'since' or 'as' in unit_string:"
                         f" {timestr}")

    return units.lower(), sincestring.lower(), remainder


def _year_zero_defaults(calendar):
    """
    Set calendar specific defaults for having year 0 or not

    Excel calendars *excel*, *excel1900*, *excel1904* start only 1900
    or above, i.e. no year 0.

    Decimal calendars *decimal*, *decimal360*, *decimal365*, *decimal366*
    have year 0 by default but might also omit it.

    Real-world calendars *standard*, *gregorian*, *julian* have no year 0.

    *proleptic_gregorian* (ISO 8601) and the idealized calendars
    *noleap*/*365_day*, *360_day*, *366_day*/*all_leap* have by default
    a year 0.

    Parameters
    ----------
    calendar : str
        One of the supported calendar names in *_noncfcalendars*

    Returns
    -------
    bool
       True if calendar includes year 0 by default

    Examples
    --------
    >>> print(_year_zero_defaults('Excel'))
    False
    >>> print(_year_zero_defaults('decimal'))
    True

    """
    calendar = calendar.lower()
    if calendar in ['standard', 'gregorian', 'julian']:
        return False
    elif calendar in ['proleptic_gregorian']:
        return True  # ISO 8601 year zero=1 BC
    elif calendar in _idealized_cfcalendars:
        return True
    elif calendar in _excelcalendars:
        return False
    elif calendar in _decimalcalendars:
        return True
    else:
        raise ValueError(f'Unknown calendar: {calendar}')


def _is_leap(year, calendar, has_year_zero=None):
    """
    Determines if a specific year in a given calendar is a leap year

    *has_year_zero* controls whether astronomical year numbering
    is used and the year zero exists. If not specified,
    calendar-specific default is assumed.

    Excel calendars *excel*, *excel1900*, *excel1904* start only in 1900
    or above, i.e. no year 0 allowed.

    Decimal calendars *decimal*, *decimal360*, *decimal365*, *decimal366*
    have year 0 by default but might omit it by setting *has_year_zero=False*.

    Parameters
    ----------
    year : int or array_like of int
        Year(s) to check if leap year
    calendar : str
        One of the supported calendar names in *_noncfcalendars*
    has_year_zero : bool, optional
        Astronomical year numbering is used, i.e. year zero exists, if True
        and possible for the given *calendar*. If *None* (default),
        calendar-specific defaults are assumed.

    Returns
    -------
    bool or array of bool
       True if year is a leap year

    Notes
    -----
    If there is no year 0 in a calendar, years -1, -5, -9, etc. are leap years.
    Year 0 is a leap year if it exists.

    Examples
    --------
    >>> years = [1900, 1904]
    >>> print(_is_leap(years, 'decimal'))
    [False, True]
    >>> print(_is_leap(years, 'Excel'))
    [True, True]

    """
    myear = input2array(year, default=1990)

    # set calendar-specific defaults for has_year_zero
    if has_year_zero is None:
        has_year_zero = _year_zero_defaults(calendar)

    if has_year_zero and (calendar in _excelcalendars):
        raise ValueError('year 0 not allowed with Excel calendars')
    if np.any(myear == 0) and (not has_year_zero):
        raise ValueError(f'year 0 does not exist in the calendar {calendar}')

    # If there is no year 0 in the calendar, years -1, -5, -9, etc.
    # are leap years. year 0 is a leap year if it exists.
    if not has_year_zero:
        myear = np.where(myear < 0, myear + 1, myear)

    if calendar in _excelcalendars:
        # Excel calendars are supposedly Julian calendars
        leap = (myear % 4) == 0
    elif calendar == 'decimal':
        leap = ( (((myear % 4) == 0) & ((myear % 100) != 0)) |
                 ((year % 400) == 0) )
    elif calendar in ['decimal360', 'decimal365']:
        leap = np.zeros_like(myear, dtype=bool)
    elif calendar == 'decimal366':
        leap = np.ones_like(myear, dtype=bool)

    oleap = array2input(leap, year)

    return oleap


def _month_lengths(year, calendar, has_year_zero=None):
    """
    Number of days of the 12 months in specific year for a given calendar

    Parameters
    ----------
    year : int
        Year to inquire
    calendar : str
        One of the supported calendar names in *_noncfcalendars*
    has_year_zero : bool, optional
        Astronomical year numbering is used, i.e. year zero exists, if True
        and possible for the given *calendar*. If *None* (default),
        calendar-specific defaults are assumed.

    Returns
    -------
    list
       Lengths of the 12 months in specified *year*

    Examples
    --------
    >>> print(_month_lengths(1990, 'decimal'))
    [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    >>> print(_month_lengths(1990, 'decimal360'))
    [30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30]

    """
    leap = _is_leap(year, calendar, has_year_zero)
    if calendar == 'decimal360':
        return _dayspermonth_360
    else:
        if leap:
            return _dayspermonth
        else:
            return _dayspermonth_leap


def _int_julian_day_from_date(year, month, day, calendar, has_year_zero=None):
    """
    Compute integer Julian Day from year, month, day, and calendar

    Allowed calendars are all in *_noncfcalendars*.

    Excel calendars are supposedly Julian calendars with other reference dates.
    Julian calendar is hence used for the Excel calendars *excel*, *excel1900*,
    *excel1904*, i.e. integer julian day is the same number of days since
    -4713-01-01 12:00:00 in the Julian calendar.

    Integer julian day in the decimal calendars *decimal*, *decimal360*,
    *decimal365*, *decimal366* is the number of days after 0000-01-01 12:00:00.

    If the has_year_zero kwarg is set to True, astronomical year numbering
    is used and the year zero exists for the decimal calendars.
    If set to False, then historical year numbering is used and the year 1 is
    preceded by year -1 and no year zero exists.
    The defaults (has_year_zero=None) uses astronomical year numbering
    for the decimal calendars.

    """
    if calendar:
        calendar = calendar.lower()
    if has_year_zero is None:
        has_year_zero = _year_zero_defaults(calendar)
    if calendar == 'decimal360':
        # return year * 360 + (month - 1) * 30 + day - 1
        return year * 360 + _cumdayspermonth_360[month-1] + day - 1
    elif calendar == 'decimal365':
        return year * 365 + _cumdayspermonth[month-1] + day - 1
    elif calendar == 'decimal366':
        return year * 366 + _cumdayspermonth_leap[month-1] + day - 1
    else:
        leap = _is_leap(year, calendar, has_year_zero=has_year_zero)
        if leap:
            jday = day + _cumdayspermonth_leap[month-1]
        else:
            jday = day + _cumdayspermonth[month-1]
        # If there is no year 0, years -1, -5, -9, etc,
        # are leap years. year zero is a leap year if it exists.
        if (year < 0) and (not has_year_zero):
            year += 1
        if calendar == 'decimal':
            # 1st term is the number of days in the last year
            # 2nd term is the number of days in each preceding non-leap year
            # last terms are the number of preceding leap years
            jday_greg = (jday + 365*(year-1) +
                         (year-1)//4 - (year-1)//100 + (year-1)//400)
            return jday_greg
        elif calendar in _excelcalendars:
            year += 4800  # add offset so -4800 is year 0.
            # 1st term is the number of days in the last year
            # 2nd term is the number of days in each preceding non-leap year
            # last terms are the number of preceding leap years
            jday_jul = jday + 365*(year-1) + (year-1)//4
            # remove offset for 87 years before -4713 (including leap days)
            jday_jul -= 31777
            return jday_jul
        else:
            raise ValueError(f'Unknown calendar: {calendar}')


# Add a datetime.timedelta to a pyjams.datetime instance. Uses
# integer arithmetic to avoid rounding errors and preserve
# microsecond accuracy.
def _add_timedelta(dt, delta):
    # extract these inputs here to avoid type conversion in the code below
    delta_microseconds = delta.microseconds
    delta_seconds = delta.seconds
    delta_days = delta.days

    # shift microseconds, seconds, days
    calendar = dt.calendar
    has_year_zero = dt.has_year_zero
    microsecond = dt.microsecond + delta_microseconds
    second = dt.second + delta_seconds
    minute = dt.minute
    hour = dt.hour
    day = dt.day
    month = dt.month
    year = dt.year

    month_length = _month_lengths(year, calendar, has_year_zero)

    # Normalize microseconds, seconds, minutes, hours.
    second += microsecond // 1000000
    microsecond = microsecond % 1000000
    minute += second // 60
    second = second % 60
    hour += minute // 60
    minute = minute % 60
    extra_days = hour // 24
    hour = hour % 24

    delta_days += extra_days

    while delta_days < 0:
        # not done compared to cftime because Excel dates > 1900 and
        # decimal dates include 1582-10-05 to 1582-10-14
        # if (year == 1582 and month == 10 and day > 14 and
        #     day + delta_days < 15):
        #     delta_days -= n_invalid_dates    # skip over invalid dates
        if (day + delta_days) < 1:
            delta_days += day
            # decrement month
            month -= 1
            if month < 1:
                month = 12
                year -= 1
                if (year == 0) and (not has_year_zero):
                    year = -1
                month_length = _month_lengths(year, calendar, has_year_zero)
            day = month_length[month-1]
        else:
            day += delta_days
            delta_days = 0

    while delta_days > 0:
        # not done compared to cftime because Excel dates > 1900 and
        # decimal dates include 1582-10-05 to 1582-10-14
        # if (year == 1582 and month == 10 and day < 5 and
        #     day + delta_days > 4):
        #     delta_days += n_invalid_dates    # skip over invalid dates
        if (day + delta_days) > month_length[month-1]:
            delta_days -= month_length[month-1] - (day - 1)
            # increment month
            month += 1
            if month > 12:
                month = 1
                year += 1
                if (year == 0) and (not has_year_zero):
                    year = 1
                month_length = _month_lengths(year, calendar, has_year_zero)
            day = 1
        else:
            day += delta_days
            delta_days = 0

    return (year, month, day, hour, minute, second, microsecond)


#
# New routines
#

def _units_defaults(calendar, has_year_zero=None):
    """
    Set calendar specific default units as 'days since reference_date'

    Day 0 of *excel* and *excel1900* starts at 1899-12-31 00:00:00.

    Day 0 of *excel1904* starts at 1903-12-31 00:00:00.

    Decimal calendars *decimal*, *decimal360*, *decimal365*, and
    *decimal366* do not need units so 0001-01-01 00:00:00 is taken.

    Day 0 of *julian*, *gregorian* and *standard* starts at
    -4713-01-01 12:00:00 if not has_year_zero, and at
    -4712-01-01 12:00:00 if has_year_zero.

    Day 0 of *proleptic_gregorian* starts at
    -4714-11-24 12:00:00 if not has_year_zero, and at
    -4713-11-24 12:00:00 if has_year_zero.

    Day 0 of *360_day*, *365_day*, *366_day*, *all_leap*, and
    *noleap* starts at 0000-01-01 12:00:00.

    Parameters
    ----------
    calendar : str
        One of the supported calendar names in *_cfcalendars* and
        *_noncfcalendars*
    has_year_zero : bool, optional
        Astronomical year numbering is used, i.e. year zero exists, if True
        and possible for the given *calendar*. If *None* (default),
        calendar-specific defaults are assumed.

    Returns
    -------
    str
       'days since reference_date' with calendar-specific reference_date

    Examples
    --------
    >>> print(_units_defaults('Excel'))
    'days since 1899-12-31 00:00:00'

    """
    calendar = calendar.lower()
    if has_year_zero is None:
        has_year_zero = _year_zero_defaults(calendar)
    if calendar in ['standard', 'gregorian', 'julian']:
        if has_year_zero:
            return 'days since -4712-01-01 12:00:00'
        else:
            return 'days since -4713-01-01 12:00:00'
    elif calendar in ['proleptic_gregorian']:
        if has_year_zero:
            return 'days since -4713-11-24 12:00:00'
        else:
            return 'days since -4714-11-24 12:00:00'
    elif calendar in _idealized_cfcalendars:
        return 'days since 0000-01-01 12:00:00'
    elif calendar in ['excel', 'excel1900']:
        return 'days since 1899-12-31 00:00:00'
    elif calendar in ['excel1904']:
        return 'days since 1903-12-31 00:00:00'
    elif calendar in _decimalcalendars:
        return 'days since 0001-01-01 00:00:00'
    else:
        raise ValueError(f'Unknown calendar: {calendar}')


def _date2decimal(date, calendar):
    """
    Decimal date from datetime object

    Parameters
    ----------
    date : datetime instance
        Instance of pyjams.datetime class
    calendar : str
        One of the decimal calendar names *decimal*, *decimal360*,
        *decimal365*, or *decimal366*

    Returns
    -------
    longdouble
       decimal date

    Examples
    --------
    >>> dt = datetime(1990, 1, 1)
    >>> dec = _date2decimal(dt, 'decimal')
    >>> print(dec)
    1990.

    """
    year     = date.year
    month    = date.month
    day      = np.longdouble(date.day)
    hour     = np.longdouble(date.hour)
    minute   = np.longdouble(date.minute)
    second   = np.longdouble(date.second)
    msecond  = np.longdouble(date.microsecond)
    calendar = calendar.lower()

    days_year = np.longdouble(365)
    diy = np.array([ [-9] + _cumdayspermonth,
                     [-9] + _cumdayspermonth_leap ], dtype=np.longdouble)
    if calendar == 'decimal':
        leap = int( (((year % 4) == 0) & ((year % 100) != 0)) |
                    ((year % 400) == 0) )
    elif calendar == 'decimal360':
        leap = 0
        days_year = np.longdouble(360)
        diy  = np.array([ [-9] + _cumdayspermonth_360,
                          [-9] + _cumdayspermonth_360 ], dtype=np.longdouble)
    elif calendar == 'decimal365':
        leap = 0
    elif calendar == 'decimal366':
        leap = 1
    fleap = np.longdouble(leap)
    tday  = diy[leap, month] + day
    thour = ( (tday - 1.) * 24. +
              hour +
              minute / 60. +
              second / 3600. +
              msecond / 3600000000. )
    out = np.longdouble(year) + thour / ((days_year + fleap) * 24.)

    return out


def _dates2decimal(dates, calendar):
    """
    Decimal dates from datetime objects

    Parameters
    ----------
    dates : datetime instance or array_like of datetime instances
        Instances of pyjams.datetime class
    calendar : str
        One of the decimal calendar names *decimal*, *decimal360*,
        *decimal365*, or *decimal366*

    Returns
    -------
    array_like of longdouble
       decimal dates

    Examples
    --------
    >>> dt = [datetime(1990, 1, 1), datetime(1991, 1, 1)]
    >>> dec = _dates2decimal(dt, 'decimal')
    >>> print(dec)
    [1990., 1991.]

    """
    mdates = input2array(dates, default=datetime(1990, 1, 1))

    # wrapper might be slow
    out = [ _date2decimal(dd, calendar) for dd in mdates ]

    out = array2input(out, dates)
    return out


def _date2absolute(date, units):
    """
    Absolute date from datetime object

    Parameters
    ----------
    date : datetime instance
        Instances of pyjams.datetime class
    units : str
        'day as %Y%m%d.%f', 'month as %Y%m.%f', or 'year as %Y.%f'

    Returns
    -------
    longdouble
       absolute date

    Examples
    --------
    >>> dt = datetime(1990, 1, 1)
    >>> dec = _date2absolute(dt, 'day as %Y%m%d.%f')
    >>> print(np.around(dec, 1))
    19900101.0

    """
    year     = date.year
    month    = date.month
    day      = np.longdouble(date.day)
    hour     = np.longdouble(date.hour)
    minute   = np.longdouble(date.minute)
    second   = np.longdouble(date.second)
    msecond  = np.longdouble(date.microsecond)

    if units == 'day as %Y%m%d.%f':
        tday  = (np.longdouble(year) * 10000. +
                 np.longdouble(month) * 100. +
                 day)
        thour = (hour +
                 minute / 60. +
                 second / 3600. +
                 msecond / 3600000000.)
        out = tday + thour / 24.
    elif units == 'month as %Y%m.%f':
        leap = int( (((year % 4) == 0) & ((year % 100) != 0)) |
                    ((year % 400) == 0) )
        dim = np.array([ [-9] + _dayspermonth,
                         [-9] + _dayspermonth_leap ], dtype=np.longdouble)
        tmonth = (np.longdouble(year) * 100. +
                  np.longdouble(month))
        thour = (day * 24. +
                 hour +
                 minute / 60. +
                 second / 3600. +
                 msecond / 3600000000.)
        out = tmonth + thour / (dim[leap, month] * 24.)
    elif units == 'year as %Y.%f':
        # same as decimal date
        out = _date2decimal(date, 'decimal')
    else:
        raise ValueError(f'Unknown absolute units: {units}')

    return out


def _dates2absolute(dates, units):
    """
    Absolute dates from datetime object

    Parameters
    ----------
    dates : datetime instance or array_like of datetime instances
        Instances of pyjams.datetime class
    units : str
        'day as %Y%m%d.%f', 'month as %Y%m.%f', or 'year as %Y.%f'

    Returns
    -------
    longdouble or array_like of longdouble
       absolute dates

    Examples
    --------
    >>> dt = [datetime(1990, 1, 1), datetime(1991, 1, 1)]
    >>> dec = _dates2absolute(dt, 'day as %Y%m%d.%f')
    >>> print(np.around(dec, 1))
    [19900101.0, 19910101.0]

    """
    mdates = input2array(dates, default=datetime(1990, 1, 1))

    # wrapper might be slow
    out = [ _date2absolute(dd, units) for dd in mdates ]

    out = array2input(out, dates)
    return out


def _decimal2date(times, calendar):
    """
    Split decimal dates into year, month, day, hour, minute, second,
    microsecond

    Parameters
    ----------
    times : float or array_like
        Decimal dates such as 1990.5102
    calendar : str
        One of the decimal calendar names *decimal*, *decimal360*,
        *decimal365*, or *decimal366*

    Returns
    -------
    tuple
       arrays of year, month, day, hour, minute, second, microsecond

    Examples
    --------
    >>> dec = [1990., 1991.5]
    >>> yr, mo, dy, hr, mi, sc, ms = _decimal2date(dec, 'decimal')
    >>> print(yr)
    [1990, 1991]
    >>> print(mo)
    [1, 7]
    >>> print(dy)
    [1, 3]

    """
    # old algorithm does decomposition on all array elements
    # should try similar to decode_dates_from_array for speed
    # where decomposition is done for the first (oldest) element only
    # and then timedeltas are added subsequently
    mtimes = input2array(times, default=1.)
    mtimes = np.array(mtimes, dtype=np.longdouble)
    calendar = calendar.lower()

    # year
    fyear = np.trunc(mtimes)
    year = fyear.astype(np.int64)
    frac_year = mtimes - fyear
    days_year = np.longdouble(365)
    diy = np.array([ [-9] + _cumdayspermonth,
                     [-9] + _cumdayspermonth_leap ])
    if calendar == 'decimal':
        leap = ( (((year % 4) == 0) & ((year % 100) != 0)) |
                 ((year % 400) == 0) ).astype(int)
        fleap = leap.astype(np.longdouble)
    elif calendar == 'decimal360':
        leap  = np.zeros_like(mtimes, dtype=int)
        fleap = np.zeros_like(mtimes, dtype=np.longdouble)
        days_year = np.longdouble(360)
        diy  = np.array([ [-9] + _cumdayspermonth_360,
                          [-9] + _cumdayspermonth_360 ])
    elif calendar == 'decimal365':
        leap  = np.zeros_like(mtimes, dtype=int)
        fleap = np.zeros_like(mtimes, dtype=np.longdouble)
    elif calendar == 'decimal366':
        leap  = np.ones_like(mtimes, dtype=int)
        fleap = np.ones_like(mtimes, dtype=np.longdouble)
    else:
        raise ValueError(f'Unknown decimal calendar: {calendar}')
    # change to microseconds to catch round-off errors,
    # i.e. cases 1 microsec less or greater than a second
    # cf. issue #187 of cftime
    # day of year in microseconds
    fhoy = frac_year * (days_year + fleap) * 86400000000.
    ihoy = np.rint(fhoy).astype(np.int64)
    ihoy = np.where(ihoy%1000000 == 1,
                    np.floor(fhoy).astype(np.int64), ihoy)
    ihoy = np.where(ihoy%1000000 == 999999,
                    np.ceil(fhoy).astype(np.int64), ihoy)
    # microsecond
    msecond = ihoy % 1000000
    ihoy    = ihoy // 1000000
    # second
    second = ihoy % 60
    ihoy   = ihoy // 60
    # minute
    minute = ihoy % 60
    ihoy   = ihoy // 60
    # hour
    hour = ihoy % 24
    ihoy = ihoy // 24
    # day and month
    idoy = ihoy + 1
    month = np.zeros_like(mtimes, dtype=np.int64)
    day   = np.zeros_like(mtimes, dtype=np.int64)
    for i in range(mtimes.size):
        ii = np.where(idoy[i] > diy[leap[i], :])[0]
        month[i] = ii[-1]
        day[i]   = idoy[i] - diy[leap[i], month[i]]

    return year, month, day, hour, minute, second, msecond


def _absolute2date(times, units):
    """
    Split date(s) in absolute date format into
    year, month, day, hour, minute, second, microsecond

    Parameters
    ----------
    times : float or array_like
        Absolute dates such as 20070102.0034722
    units : str
        'day as %Y%m%d.%f', 'month as %Y%m.%f', or 'year as %Y.%f'

    Returns
    -------
    tuple
       arrays of year, month, day, hour, minute, second, microsecond

    Examples
    --------
    >>> absolut = [20070102.0034722, 20070102.0069444]
    >>> yr, mo, dy, hr, mi, sc, ms = _absolute2date(absolut,
    ...                                             'day as %Y%m%d.%f')
    >>> print(yr)
    [2007, 2007]
    >>> print(mi)
    [5, 10]

    """
    # old algorithm does decomposition on all array elements
    # should try similar to decode_dates_from_array for speed
    # where decomposition is done for the first (oldest) element only
    # and then timedeltas are added subsequently
    mtimes = input2array(times, default=10101.)
    mtimes = np.array(mtimes, dtype=np.longdouble)

    if units == 'day as %Y%m%d.%f':
        # change to microseconds to catch round-off errors,
        # i.e. cases 1 microsec less or greater than a second
        # cf. issue #187 of cftime
        # day of year in microseconds
        fhoy = mtimes * 86400000000.
        ihoy = np.rint(fhoy).astype(np.int64)
        ihoy = np.where(ihoy%1000000 == 1,
                        np.floor(fhoy).astype(np.int64), ihoy)
        ihoy = np.where(ihoy%1000000 == 999999,
                        np.ceil(fhoy).astype(np.int64), ihoy)
        # microsecond
        msecond = ihoy % 1000000
        ihoy    = ihoy // 1000000
        # second
        second = ihoy % 60
        ihoy   = ihoy // 60
        # minute
        minute = ihoy % 60
        ihoy   = ihoy // 60
        # hour
        hour = ihoy % 24
        ihoy = ihoy // 24
        # day
        day = ihoy % 100
        ihoy = ihoy // 100
        # month
        month = ihoy % 100
        ihoy = ihoy // 100
        # year
        year = ihoy
    elif units == 'month as %Y%m.%f':
        fmo     = mtimes % 1.  # month fraction
        # month
        mtimes -= fmo
        month   = np.rint(mtimes % 100.).astype(np.int64)
        # year
        mtimes -= month
        year    = np.rint(mtimes / 100.).astype(np.int64)
        # day of month in microseconds
        leap    = np.where((((year % 4) == 0) & ((year % 100) != 0)) |
                           ((year % 400) == 0), 1, 0)
        dim     = np.array([ [-9] + _dayspermonth,
                             [-9] + _dayspermonth_leap ])
        fhoy = dim[(leap, month)] * fmo * 86400000000.
        ihoy = np.rint(fhoy).astype(np.int64)
        ihoy = np.where(ihoy%1000000 == 1,
                        np.floor(fhoy).astype(np.int64), ihoy)
        ihoy = np.where(ihoy%1000000 == 999999,
                        np.ceil(fhoy).astype(np.int64), ihoy)
        # microsecond
        msecond = ihoy % 1000000
        ihoy    = ihoy // 1000000
        # second
        second = ihoy % 60
        ihoy   = ihoy // 60
        # minute
        minute = ihoy % 60
        ihoy   = ihoy // 60
        # hour
        hour = ihoy % 24
        ihoy = ihoy // 24
        # day
        day = ihoy
        # mtimes  = dim[(leap, month)] * fmo
        # fdy     = mtimes % 1.  # day fraction
        # mtimes -= fdy
        # day     = np.rint(mtimes % 100.).astype(np.int64)
        # # hour
        # secs    = fdy * 86400.
        # fsecs   = secs % 1.  # second fraction
        # secs    = np.rint(secs)
        # hour    = np.floor(secs / 3600.).astype(np.int64)
        # # minute
        # secs   -= 3600. * hour
        # minute  = np.floor(secs / 60.).astype(np.int64)
        # # second
        # secs   -= 60. * minute
        # second  = np.rint(secs).astype(np.int64)
        # # millisecond
        # msecond  = np.rint(fsecs * 1000000.).astype(np.int64)
    elif units == 'year as %Y.%f':
        # same as decimal date
        year, month, day, hour, minute, second, msecond = _decimal2date(
            times, 'decimal')
    else:
        raise ValueError(f'Unknown absolute units: {units}')

    return year, month, day, hour, minute, second, msecond


#
# Main routines and class
#

def date2num(dates, units='', calendar=None, has_year_zero=None,
             format='', timesep=' ', fr=False):
    """
    Return numeric time values given datetime objects or strings

    The units of the numeric time values are described by the
    *units* and *calendar* keywords for CF-conform calendars, i.e.
    *standard*, *gregorian*, *julian*, *proleptic_gregorian*,
    *360_day*, *365_day*, *366_day*, *noleap*, *all_leap*.
    See http://cfconventions.org/cf-conventions/cf-conventions#calendar
    These times will be passed to cftime.num2date.

    Standard *units* are used for the non-CF-conform calendars
    *excel*, *excel1900*, *excel1904*, and
    *decimal*, *decimal360*, *decimal365*, *decimal366*,
    and given *units* will hence be ignored.

    Parameters
    ----------
    dates : datetime instance or str, or array_like of datetime or str
        datetime objects or strings with string representations.
        datetime objects can either be Python datetime.datetime,
        cf.datetime, or pyjams.datetime objects. If *dates* are strings,
        then *format* keyword is relevant.
    units : str, optional
        Units string such as 'seconds since 1900-01-01 00:00:00' or
        'day as %Y%m%d.%f'. Standard units corresponding to days after
        day 0 of a given calendar will be used if omitted, i.e. assuming
        Julian day ordinals.

        In the form *time_units since reference_time*,
        *time_units* can be days, hours, minutes, seconds, milliseconds,
        or microseconds. *reference_time* is the time origin.
        *months since* is allowed only for the *360_day* calendar
        and *common_years since* is allowed only for the *365_day* calendar.

        There are currently only three valid forms for *time_units as format*:
        'day as %Y%m%d.%f', 'month as %Y%m.%f', 'year as %Y.%f'. The latter is
        the same as 'calendar=decimal'. *calendar='decimal' will be set in
        case units *time_units as format*.
    calendar : str, optional
        One of the support calendars, i.e. the CF-conform calendars
        *standard*, *gregorian*, *julian*, *proleptic_gregorian*,
        *360_day*, *365_day*, *366_day*, *noleap*, *all_leap*,
        as well as the non-CF-conform calendars
        *excel*, *excel1900*, *excel1904*, and
        *decimal*, *decimal360*, *decimal365*, *decimal366*.
        *standard* will be taken by default, which is a mixed
        Julian/Gregorian calendar.
    has_year_zero : bool, optional
        Astronomical year numbering is used and the year zero exists, if set to
        True. If set to False for real-world calendars, then historical year
        numbering is used and the year 1 is preceded by year -1 and no year
        zero exists.
        The defaults are set to conform with CF version 1.9 conventions, i.e.
        False for 'standard', 'gregorian', and 'julian', True
        for 'proleptic_gregorian' (ISO 8601), and True for the idealized
        calendars 'noleap'/'365_day', '360_day', 366_day'/'all_leap'.
        Excel calendars *excel*, *excel1900*, *excel1904* start only 1900
        or above so *has_year_zero* is always False.
        Decimal calendars *decimal*, *decimal360*, *decimal365*, *decimal366*
        have always year 0, i.e. *has_year_zero* is True.
    format : str, optional
        If *dates* are strings, then *format* is the Python
        datetime.strftime/strptime format string if given. If empty (default),
        then the routine pyjams.date2date will be used, which converts between
        formats '%Y-%m-%d %H:%M:%S', '%d.%m.%Y %H:%M:%S', and '%m/%d/%Y
        %H:%M:%S' (called English *YYYY-MM-DD hh:mm:ss*, standard
        *DD.MM.YYYY hh:mm:ss*, and American *MM/DD/YYYY hh:mm:ss* in
        pyjams.date2date), where times can be partial or missing.
    timesep : str, optional
        Separator string between date and time used by pyjams.date2date if
        if *dates* are strings and *format* is empty (default: ' ')
    fr : bool, optional
        If True, pyjams.date2date will interpret input dates with '/'
        separators not as the American format '%m/%d/%Y %H:%M:%S' but the
        French way as '%d/%m/%Y %H:%M:%S', if *dates* are strings and
        *format* is empty

    Returns
    -------
    array_like
       numeric time values

    Examples
    --------
    >>> idates = ['2000-01-05 12:30:15', '1810-04-24 16:15:10',
    ...           '1630-07-15 10:20:40', '1510-09-20 14:35:50',
    ...           '1271-03-18 19:41:34', '0619-08-27 11:08:37',
    ...           '0001-01-01 12:00:00']
    >>> decimal = date2num(idates, calendar='decimal')
    >>> num2date(decimal, calendar='decimal', format='%Y-%m-%d %H:%M:%S')
    ['2000-01-05 12:30:15',
     '1810-04-24 16:15:10',
     '1630-07-15 10:20:40',
     '1510-09-20 14:35:50',
     '1271-03-18 19:41:34',
     '0619-08-27 11:08:37',
     '0001-01-01 12:00:00']

    """
    date0 = np.ravel(dates)[0]
    if isinstance(date0, str):
        if format:
            iform = format
        else:
            iform = '%Y-%m-%d %H:%M:%S'
        default = cf.real_datetime(1990, 1, 1).strftime(iform)
    else:
        default = cf.real_datetime(1990, 1, 1)
    mdates = input2array(dates, default=default)

    if calendar:
        calendar = calendar.lower()
    else:
        calendar = 'standard'
    if (calendar not in _cfcalendars) and (calendar not in _noncfcalendars):
        raise ValueError(f'Unknown calendar: {calendar}')
    if not units:
        units = _units_defaults(calendar)

    # transform strings to datetime objects
    # only possible for year > 0
    isstr = all([ isinstance(dd, str) for dd in mdates ])
    if isstr:
        if format:
            iform = format
        else:
            mdates = date2date(mdates, format='en', full=True,
                               timesep=timesep, fr=fr)
            iform = '%Y-%m-%d %H:%M:%S'
        mmdates = [ cf.real_datetime.strptime(dd, iform)
                    for dd in mdates ]
        mdates = input2array(mmdates, default=cf.real_datetime(1990, 1, 1))
    else:
        mdates = input2array(dates, default=cf.real_datetime(1990, 1, 1))

    # input calendar
    date0 = mdates[0]
    try:
        icalendar = date0.calendar
    except AttributeError:
        # must be Python datetime object
        icalendar = 'proleptic_gregorian'

    # check if we can parse to cftime
    if (icalendar in _cfcalendars) and (calendar in _cfcalendars):
        iscf = True
    elif (icalendar in _noncfcalendars) or (calendar in _noncfcalendars):
        iscf = False
    else:
        raise ValueError(f'Unknown calendar: {icalendar} or {calendar}')

    unit, sincestr, remainder = _datesplit(units)
    if sincestr == 'as':
        iscf = False
        calendar = ''

    # use cftime.date2num if possible
    if iscf:
        out = cf.date2num(mdates, units, calendar=calendar,
                          has_year_zero=has_year_zero)

    if sincestr == 'as':
        if units not in ['day as %Y%m%d.%f', 'month as %Y%m.%f',
                         'year as %Y.%f']:
            raise ValueError(f'Absolute date format unknown: {units}')
        out = _dates2absolute(mdates, units)

    # use cftime.date2num with Excel
    if calendar in _excelcalendars:
        cfcalendar = 'julian'
        # cf.real_datetime has no to_tuple() method yet
        # cfdates = [ cf.datetime(*dt.to_tuple(), calendar=cfcalendar)
        #             for dt in mdates ]
        cfdates = [ cf.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute,
                                dt.second, dt.microsecond, calendar=cfcalendar)
                    for dt in mdates ]
        out = cf.date2num(cfdates, units, calendar=cfcalendar,
                          has_year_zero=has_year_zero)

    # no cftime.num2date possible
    if calendar in _decimalcalendars:
        out = _dates2decimal(mdates, calendar)

    out = array2input(out, dates)
    return out


def date2dec(*args, **kwargs):
    """
    Wrapper for :func:`date2num`
    """
    return date2num(*args, **kwargs)


def num2date(times, units='', calendar='standard',
             only_use_cftime_datetimes=True,
             only_use_python_datetimes=False,
             has_year_zero=None,
             format='', return_arrays=False):
    """
    Return datetime objects given numeric time values

    The units of the numeric time values are described by the
    *units* and *calendar* keywords for CF-conform calendars, i.e.
    *standard*, *gregorian*, *julian*, *proleptic_gregorian*,
    *360_day*, *365_day*, *366_day*, *noleap*, *all_leap*.
    See http://cfconventions.org/cf-conventions/cf-conventions#calendar
    These times will be passed to cftime.num2date.

    Standard *units* are used for the non-CF-conform calendars
    *excel*, *excel1900*, *excel1904*, and
    *decimal*, *decimal360*, *decimal365*, *decimal366*,
    and given *units* will hence be ignored.

    Parameters
    ----------
    times : float or array_like
        Numeric time values
    units : str, optional
        Units string such as 'seconds since 1900-01-01 00:00:00' or
        'day as %Y%m%d.%f'. Standard units corresponding to days after
        day 0 of a given calendar will be used if omitted, i.e. assuming
        Julian day ordinals.

        In the form *time_units since reference_time*,
        *time_units* can be days, hours, minutes, seconds, milliseconds,
        or microseconds. *reference_time* is the time origin.
        *months since* is allowed only for the *360_day* calendar
        and *common_years since* is allowed only for the *365_day* calendar.

        There are currently only three valid forms for *time_units as format*:
        'day as %Y%m%d.%f', 'month as %Y%m.%f', 'year as %Y.%f'. The latter is
        the same as 'calendar=decimal'. *calendar='decimal' will be set in
        case units *time_units as format*.
    calendar : str, optional
        One of the support calendars, i.e. the CF-conform calendars
        *standard*, *gregorian*, *julian*, *proleptic_gregorian*,
        *360_day*, *365_day*, *366_day*, *noleap*, *all_leap*,
        as well as the non-CF-conform calendars
        *excel*, *excel1900*, *excel1904*, and
        *decimal*, *decimal360*, *decimal365*, *decimal366*.
        *standard* will be taken by default, which is a mixed
        Julian/Gregorian calendar.
    only_use_cftime_datetimes : bool, optional
        pyjams.datetime and cftime.datetime objects are returned by default.
        Only if only_use_cftime_datetimes is set to False (default: True)
        and only_use_python_datetimes is set to True then Python
        datetime.datetime objects will be returned where possible.
    only_use_python_datetimes : bool, optional
        pyjams.datetime and cftime.datetime objects are returned by default.
        Only if only_use_python_datetimes is set to True (default: False)
        and only_use_cftime_datetimes is set to False then Python
        datetime.datetime objects will be returned where possible.
    has_year_zero : bool, optional
        Astronomical year numbering is used and the year zero exists, if set to
        True. If set to False for real-world calendars, then historical year
        numbering is used and the year 1 is preceded by year -1 and no year
        zero exists.
        The defaults are set to conform with CF version 1.9 conventions, i.e.
        False for 'standard', 'gregorian', and 'julian', True
        for 'proleptic_gregorian' (ISO 8601), and True for the idealized
        calendars 'noleap'/'365_day', '360_day', 366_day'/'all_leap'.
        Excel calendars *excel*, *excel1900*, *excel1904* start only 1900
        or above so *has_year_zero* is always False.
        Decimal calendars *decimal*, *decimal360*, *decimal365*, *decimal366*
        have always year 0, i.e. *has_year_zero* is True.
    format : str, optional
        If format string is given than a string representation of the
        datetime objects will be returned.
    return_arrays : bool, optional
        If True, then return a tuple with individual arrays for
        year, month, day, hour, minute, second, microsecond

    Returns
    -------
    array_like
       datetime instances or string representations of datetime objects, or
       tuple with individual arrays for year, month, day, hour, minute,
       second, microsecond

    Examples
    --------
    >>> idates = ['2000-01-05 12:30:15', '1810-04-24 16:15:10',
    ...           '1630-07-15 10:20:40', '1510-09-20 14:35:50',
    ...           '1271-03-18 19:41:34', '0619-08-27 11:08:37',
    ...           '0001-01-01 12:00:00']
    >>> decimal = date2num(idates, calendar='decimal')
    >>> num2date(decimal, calendar='decimal', format='%Y-%m-%d %H:%M:%S')
    ['2000-01-05 12:30:15',
     '1810-04-24 16:15:10',
     '1630-07-15 10:20:40',
     '1510-09-20 14:35:50',
     '1271-03-18 19:41:34',
     '0619-08-27 11:08:37',
     '0001-01-01 12:00:00']

    """
    if format and return_arrays:
        raise ValueError('Keywords format and return_arrays mutually'
                         ' exclusive')
    if calendar:
        calendar = calendar.lower()
    else:
        calendar = 'standard'
    # check if we can parse to cftime
    if calendar in _cfcalendars:
        iscf = True
    elif calendar in _noncfcalendars:
        iscf = False
    else:
        raise ValueError(f'Unknown calendar: {calendar}')
    if not units:
        units = _units_defaults(calendar)
    unit, sincestr, remainder = _datesplit(units)
    if sincestr == 'as':
        iscf = False
        calendar = ''

    # use cftime.num2date if possible
    if iscf:
        # print(times, units, calendar)
        out = cf.num2date(
            times, units, calendar=calendar,
            only_use_cftime_datetimes=only_use_cftime_datetimes,
            only_use_python_datetimes=only_use_python_datetimes,
            has_year_zero=has_year_zero)

    # cdo absolute time format
    if sincestr == 'as':
        if units not in ['day as %Y%m%d.%f', 'month as %Y%m.%f',
                         'year as %Y.%f']:
            raise ValueError(f'Absolute date format unknown: {units}')
        # old algorithm does decomposition on all array elements
        # should try similar to decode_dates_from_array for speed
        # where decomposition is done for the first (oldest) element only
        # and then timedeltas are added subsequently
        year, month, day, hour, minute, second, microsecond = (
            _absolute2date(times, units))

        # shortcut
        if return_arrays:
            year        = array2input(year, times)
            month       = array2input(month, times)
            day         = array2input(day, times)
            hour        = array2input(hour, times)
            minute      = array2input(minute, times)
            second      = array2input(second, times)
            microsecond = array2input(microsecond, times)
            return year, month, day, hour, minute, second, microsecond

        out = np.empty_like(year, dtype=object)
        if ( (not only_use_cftime_datetimes) and
             only_use_python_datetimes and (year.min() > 0) ):
            for i in range(year.size):
                out[i] = cf.real_datetime(year[i], month[i], day[i],
                                          hour[i], minute[i], second[i],
                                          microsecond[i])
        else:
            for i in range(year.size):
                out[i] = datetime(year[i], month[i], day[i],
                                  hour[i], minute[i], second[i],
                                  microsecond[i],
                                  calendar='decimal',
                                  has_year_zero=has_year_zero)

    # use cftime.num2date for Excel but return pyjams.datetime
    if calendar in _excelcalendars:
        cfcalendar = 'julian'
        cfunits = _units_defaults(calendar)
        cfdates = cf.num2date(
            times, cfunits, calendar=cfcalendar,
            only_use_cftime_datetimes=only_use_cftime_datetimes,
            only_use_python_datetimes=False,
            has_year_zero=has_year_zero)

        # shortcut
        if format:
            # Assure 4 digit years on all platforms
            # see https://bugs.python.org/issue32195
            iform = format
            if '%Y' in format:
                format04 = format.replace('%Y', '%04Y')
                dttest = cfdates[0].strftime(format04)
                if '4Y' in dttest:
                    iform = format
                else:
                    iform = format04
            out = [ dt.strftime(iform) for dt in cfdates ]
            out = array2input(out, times)
            return out

        # cf.real_datetime has no to_tuple() method yet
        # out = [ datetime(*dt.to_tuple(), calendar=calendar)
        out = [ datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute,
                         dt.second, dt.microsecond, calendar=calendar)
                for dt in cfdates ]

    # no cftime.num2date possible
    if calendar in _decimalcalendars:
        # old algorithm does decomposition on all array elements
        # should try similar to decode_dates_from_array for speed
        # where decomposition is done for the first (oldest) element only
        # and then timedeltas are added subsequently
        year, month, day, hour, minute, second, microsecond = (
            _decimal2date(times, calendar))

        # shortcut
        if return_arrays:
            year        = array2input(year, times)
            month       = array2input(month, times)
            day         = array2input(day, times)
            hour        = array2input(hour, times)
            minute      = array2input(minute, times)
            second      = array2input(second, times)
            microsecond = array2input(microsecond, times)
            return year, month, day, hour, minute, second, microsecond

        out = np.empty_like(year, dtype=object)
        if ( (not only_use_cftime_datetimes) and
             only_use_python_datetimes and (year.min() > 0) ):
            for i in range(year.size):
                out[i] = cf.real_datetime(year[i], month[i], day[i],
                                          hour[i], minute[i], second[i],
                                          microsecond[i])
        else:
            for i in range(year.size):
                out[i] = datetime(year[i], month[i], day[i],
                                  hour[i], minute[i], second[i],
                                  microsecond[i],
                                  calendar=calendar,
                                  has_year_zero=has_year_zero)

    if return_arrays:
        # cf.real_datetime has no to_tuple() method yet
        # out = np.array([ dt.to_tuple() for dt in out ])
        out = np.array([ (dt.year, dt.month, dt.day, dt.hour, dt.minute,
                         dt.second, dt.microsecond) for dt in out ])
        year        = array2input(out[:, 0], times)
        month       = array2input(out[:, 1], times)
        day         = array2input(out[:, 2], times)
        hour        = array2input(out[:, 3], times)
        minute      = array2input(out[:, 4], times)
        second      = array2input(out[:, 5], times)
        microsecond = array2input(out[:, 6], times)
        return year, month, day, hour, minute, second, microsecond
    else:
        if format:
            # Assure 4 digit years on all platforms
            # see https://bugs.python.org/issue32195
            iform = format
            if '%Y' in format:
                format04 = format.replace('%Y', '%04Y')
                dttest = out[0].strftime(format04)
                if '4Y' in dttest:
                    iform = format
                else:
                    iform = format04
            out = [ dt.strftime(iform) for dt in out ]

        out = array2input(out, times)
        return out


def dec2date(*args, **kwargs):
    """
    Wrapper for :func:`num2date`
    """
    return num2date(*args, **kwargs)


# Could not inherit from cf.datetime because all methods are read-only,
# so had to re-code all methods, even identical ones
class datetime(object):
    """
    This class mimics cftime.datetime but for non-CF-conform calendars

    The cftime.datetime class mimics itself datetime.datetime but
    supports calendars other than the proleptic Gregorian calendar.

    This class supports timedelta operations by overloading +/-, and
    comparisons with other instances using the same calendar.

    Current supported calendars are *excel*, *excel1900*, *excel1904*,
    and *decimal*, *decimal360*, *decimal365*, *decimal366*. Excel
    calendars are supposedly Julian calendars with other reference dates.
    Day 0 of *excel*/*excel1900* starts at 1899-12-31 00:00:00 and day 0
    of *excel1904* (old Lotus date) starts at 1903-12-31 00:00:00.
    Decimal calendars have the form "%Y.%f", where the fractional year
    assumes leap years as the proleptic Gregorian calendar, or fixed 360,
    365, or 366 days per year.

    If the has_year_zero keyword argument is set to True, astronomical year
    numbering is used and the year zero exists, which is the default for the
    decimal calendars. The keyword will be ignored for Excel calendars.

    The class has the methods isoformat, strftime, timetuple, replace,
    dayofwk, dayofyr, daysinmonth, __repr__, __format__, __add__, __sub__,
    __str__, and comparison methods.

    The default format of the string produced by strftime is controlled by
    self.format (default %Y-%m-%d %H:%M:%S).

    """
    # Python's datetime.datetime uses the proleptic Gregorian
    # calendar. This boolean is used to decide whether a
    # cftime.datetime instance can be converted to
    # datetime.datetime.
    def __init__(self, year, month, day,
                 hour=0, minute=0, second=0, microsecond=0,
                 dayofwk=-1, dayofyr=-1,
                 calendar='decimal', has_year_zero=None):
        """
        Initialise new datetime instance

        """
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.microsecond = microsecond
        self._dayofwk = dayofwk
        self._dayofyr = dayofyr
        self.tzinfo = None
        if calendar:
            self.calendar = calendar.lower()
        else:
            self.calendar = calendar
        if self.calendar in _cfcalendars:
            raise ValueError(f'Use cftime.datetime for CF-conform'
                             f' calendars: {self.calendar}')
        if has_year_zero is None:
            self.has_year_zero = _year_zero_defaults(calendar)
        else:
            self.has_year_zero = has_year_zero
        if self.calendar and (self.calendar not in _noncfcalendars):
            raise ValueError(f'Unknown calendar: {self.calendar}')
        if self.calendar in _excelcalendars:
            self.datetime_compatible = True
        else:
            self.datetime_compatible = False
        self.assert_valid_date()

    def assert_valid_date(self):
        """
        Check that datetime is a valid date for given calendar

        """
        # year
        if not self.has_year_zero:
            if self.year == 0:
                raise ValueError("Invalid year provided in {0!r}".format(self))
        # Comment next block to allow negative days with Excel calendars,
        # which does not exist in Excel
        # if ( ((self.calendar == 'excel') or (self.calendar == 'excel1900')) and
        #      self.year < 1900):
        #     raise ValueError('Year must be >= 1900 for Excel dates')
        # if ( (self.calendar == 'excel1904') and self.year < 1904):
        #     raise ValueError('Year must be >= 1904 for Excel1904 dates')

        # month
        if (self.month < 1) or (self.month > 12):
            raise ValueError("Invalid month provided in {0!r}".format(self))

        # day
        month_length = _month_lengths(self.year, self.calendar,
                                      self.has_year_zero)
        if (self.day < 1) or (self.day > month_length[self.month-1]):
            raise ValueError(
                "Invalid day number provided in {0!r}".format(self))

        # hour
        if (self.hour < 0) or (self.hour > 23):
            raise ValueError("Invalid hour provided in {0!r}".format(self))

        # minute
        if (self.minute < 0) or (self.minute > 59):
            raise ValueError("Invalid minute provided in {0!r}".format(self))

        # second
        if (self.second < 0) or (self.second > 59):
            raise ValueError("Invalid second provided in {0!r}".format(self))

        # microsecond
        if (self.microsecond) < 0 or (self.microsecond > 999999):
            raise ValueError(
                "Invalid microsecond provided in {0!r}".format(self))

    def change_calendar(self, calendar, has_year_zero=None):
        return NotImplemented

    def dayofwk(self):
        """
        Day of the week

        Identical to cftime.datetime

        """
        if (self._dayofwk < 0) and self.calendar:
            jd = self.toordinal()
            dayofwk = (jd + 1) % 7
            # convert to ISO 8601 (0 = Monday, 6 = Sunday), like python
            # datetime
            dayofwk -= 1
            if dayofwk == -1:
                dayofwk = 6
            # cache results for dayofwk
            self._dayofwk = dayofwk
            return dayofwk
        else:
            return self._dayofwk

    def dayofyr(self):
        """
        Day of year

        """
        if (self._dayofyr < 0) and self.calendar:
            if self.calendar == 'decimal360':
                # dayofyr = (self.month - 1) * 30 + self.day
                dayofyr = _cumdayspermonth_360[self.month-1] + self.day
            else:
                if _is_leap(self.year, self.calendar,
                            has_year_zero=self.has_year_zero):
                    dayofyr = _cumdayspermonth_leap[self.month-1] + self.day
                else:
                    dayofyr = _cumdayspermonth[self.month-1] + self.day
            # cache results for dayofyr
            self._dayofyr = dayofyr
            return dayofyr
        else:
            return self._dayofyr

    def daysinmonth(self):
        """
        Number of days in current month

        """
        if self.calendar == 'decimal360':
            # return 30
            return _dayspermonth_360[self.month-1]
        else:
            if _is_leap(self.year, self.calendar,
                        has_year_zero=self.has_year_zero):
                return _dayspermonth_leap[self.month-1]
            else:
                return _dayspermonth[self.month-1]

    def format(self):
        """
        Standard date representation

        Identical to cftime.datetime

        """
        return '%Y-%m-%d %H:%M:%S'

    def fromordinal(jday, calendar='decimal', has_year_zero=None):
        return NotImplemented

    def isoformat(self, sep='T', timespec='auto'):
        """
        ISO date representation

        Identical to cftime.datetime

        """
        second = ":%02i" % self.second
        if ( ((timespec == 'auto') and self.microsecond) or
             (timespec == 'microseconds') ):
            second += ".%06i" % self.microsecond
        if timespec == 'milliseconds':
            millisecs = self.microsecond / 1000
            second += ".%03i" % millisecs
        if timespec in ['auto', 'microseconds', 'milliseconds']:
            return ("%04i-%02i-%02i%s%02i:%02i%s" %
                    (self.year, self.month, self.day, sep,
                     self.hour, self.minute, second))
        elif timespec == 'seconds':
            return ("%04i-%02i-%02i%s%02i:%02i:%02i" %
                    (self.year, self.month, self.day, sep,
                     self.hour, self.minute, self.second))
        elif timespec == 'minutes':
            return ("%04i-%02i-%02i%s%02i:%02i" %
                    (self.year, self.month, self.day, sep,
                     self.hour, self.minute))
        elif timespec == 'hours':
            return ("%04i-%02i-%02i%s%02i" %
                    (self.year, self.month, self.day, sep, self.hour))
        else:
            raise ValueError('illegal timespec')

    def replace(self, **kwargs):
        """
        Return datetime with new specified fields

        Identical to cftime.datetime

        """
        args = {"year": self.year,
                "month": self.month,
                "day": self.day,
                "hour": self.hour,
                "minute": self.minute,
                "second": self.second,
                "microsecond": self.microsecond,
                "has_year_zero": self.has_year_zero,
                "calendar": self.calendar}

        if 'dayofyr' in kwargs or 'dayofwk' in kwargs:
            raise ValueError('Replacing the dayofyr or dayofwk of a datetime'
                             ' is not supported.')

        if 'calendar' in kwargs:
            raise ValueError('Replacing the calendar of a datetime is '
                             'not supported.')

        # if attempting to set year to zero, also set has_year_zero=True
        # (issue #248)
        if 'year' in kwargs:
            if (kwargs['year'] == 0) and ('has_year_zero' not in kwargs):
                kwargs['has_year_zero'] = True

        for name, value in kwargs.items():
            args[name] = value

        return self.__class__(**args)

    def strftime(self, format=None):
        """
        Return a string representing the date, controlled by an explicit format
        string

        For a complete list of formatting directives, see section 'strftime()
        and strptime() Behavior' in the base Python documentation.

        Identical to cftime.datetime

        """
        if format is None:
            format = self.format()
        return _strftime(self, format)

    def timetuple(self):
        """
        Return a time.struct_time such as returned by time.localtime()

        The DST flag is -1. d.timetuple() is equivalent to
        time.struct_time((d.year, d.month, d.day, d.hour, d.minute,
        d.second, d.weekday(), yday, dst)), where yday is the
        day number within the current year starting with 1 for January 1st.

        Identical to cftime.datetime

        """
        return ptime.struct_time((self.year, self.month, self.day, self.hour,
                                  self.minute, self.second, self.dayofwk(),
                                  self.dayofyr(), -1))

    def toordinal(self, fractional=False):
        """
        Julian day (integer) ordinal

        Day 0 starts at noon January 1 of the year -4713 for the
        Excel calendars.

        Day 0 starts at noon on January 1 of the year zero for
        the decimal calendars.

        If fractional=True, fractional part of day is included (default
        False).

        """
        ijd = _int_julian_day_from_date(
            self.year, self.month, self.day, self.calendar,
            has_year_zero=self.has_year_zero)
        if fractional:
            # At this point ijd is an integer representing noon UTC on the
            # given year, month, day.
            # Compute fractional day from hour, minute, second, microsecond
            fracday = ( self.hour / np.array(24., np.longdouble) +
                        self.minute / np.array(1440., np.longdouble) +
                        (self.second +
                         self.microsecond / (np.array(1.e6, np.longdouble))) /
                        np.array(86400., np.longdouble) )
            return ijd - 0.5 + fracday
        else:
            return ijd

    def to_tuple(self):
        """
        Tuple with year, month, day, hour, minute, second, microsecond

        Identical to cftime.datetime

        """
        return (self.year, self.month, self.day, self.hour, self.minute,
                self.second, self.microsecond)

    def _add_timedelta(self, other):
        return self.__add__(other)

    def _getstate(self):
        """
        return args and kwargs needed to create class instance

        Identical to cftime.datetime

        """
        args = (self.year, self.month, self.day)
        kwargs = {'hour': self.hour,
                  'minute': self.minute,
                  'second': self.second,
                  'microsecond': self.microsecond,
                  'dayofwk': self._dayofwk,
                  'dayofyr': self._dayofyr,
                  'calendar': self.calendar,
                  'has_year_zero': self.has_year_zero}
        return args, kwargs

    def _to_real_datetime(self):
        """
        Extended Python datetime class

        Extra attributes are  dayofwk, dayofyr, and daysinmonth.

        Identical to cftime.datetime

        """
        return cf.real_datetime(self.year, self.month, self.day,
                                self.hour, self.minute, self.second,
                                self.microsecond)

    def __add__(self, other):
        """
        Add timedelta to datetime

        """
        if isinstance(self, datetime) and isinstance(other, timedelta):
            dt = self
            calendar = self.calendar
            has_year_zero = self.has_year_zero
            delta = other
        elif isinstance(self, timedelta) and isinstance(other, datetime):
            dt = other
            calendar = other.calendar
            has_year_zero = other.has_year_zero
            delta = self
        else:
            return NotImplemented
        if calendar == 'decimal360':
            year, month, day, hour, minute, second, microsecond, _1, _2 = (
                cf.add_timedelta_360_day(dt, delta))
        else:
            year, month, day, hour, minute, second, microsecond = (
                _add_timedelta(dt, delta))
        return datetime(year, month, day,
                        hour, minute, second, microsecond,
                        calendar=dt.calendar, has_year_zero=dt.has_year_zero)

    def __cmp__(self, other):
        """
        Compare two datetime instances

        """
        dt = self
        if isinstance(other, datetime):
            dt_other = other
            # comparing two datetime instances
            if ( (dt.calendar == dt_other.calendar) and
                 (dt.has_year_zero == dt_other.has_year_zero) ):
                return dt.to_tuple() == dt_other.to_tuple()
            else:
                return (dt.toordinal(fractional=True) ==
                        dt_other.toordinal(fractional=True))
        else:
            return NotImplemented

    def __format__(self, format):
        """
        Return a string representing the date, controlled by an explicit format
        string

        For a complete list of formatting directives, see section 'strftime()
        and strptime() Behavior' in the base Python documentation.

        Identical to cftime.datetime

        """
        # the string format "{t_obj}".format(t_obj=t_obj)
        # without an explicit format gives an empty string (format='')
        # so set this to None to get the default strftime behaviour
        if not format:
            format = None
        return self.strftime(format)

    def __hash__(self):
        """
        Identical to cftime.datetime

        """
        try:
            d = self._to_real_datetime()
        except ValueError:
            return hash(self.timetuple())
        return hash(d)

    def __reduce__(self):
        """
        Special method that allows instance to be pickled

        Identical to cftime.datetime

        """
        args, kwargs = self._getstate()
        date_type = type(self)
        return (_create_datetime, (date_type, args, kwargs))

    def __repr__(self):
        """
        String representation

        """
        return ("{0}.{1}({2}, {3}, {4}, {5}, {6}, {7}, {8},"
                " calendar={9}, has_year_zero={10})".format(
                    'pyjams',
                    self.__class__.__name__,
                    self.year, self.month, self.day,
                    self.hour, self.minute, self.second,
                    self.microsecond, self.calendar, self.has_year_zero))

    def __richcmp__(self, other):
        """
        Compare two datetime instances

        """
        return self.__cmp__(other)

    def __str__(self):
        """
        ISO date representation

        Identical to cftime.datetime

        """
        return self.isoformat(' ')

    def __sub__(self, other):
        """
        Substract timedelta or datetime from the datetime instance

        """
        if isinstance(self, datetime):  # left arg is a datetime instance
            dt = self
            if isinstance(other, datetime):
                # datetime - datetime
                if dt.calendar != other.calendar:
                    raise TypeError("Cannot compute the time difference"
                                    " between dates with different calendars")
                if dt.calendar == "":
                    raise TypeError("Cannot compute the time difference"
                                    " between dates that are not"
                                    " calendar-aware")
                if dt.has_year_zero != other.has_year_zero:
                    raise TypeError("Cannot compute the time difference"
                                    " between dates with different year zero"
                                    " conventions")
                ordinal_self = self.toordinal()  # julian day
                ordinal_other = other.toordinal()
                days = ordinal_self - ordinal_other
                seconds_self = dt.second + 60 * dt.minute + 3600 * dt.hour
                seconds_other = (other.second + 60 * other.minute +
                                 3600 * other.hour)
                seconds = seconds_self - seconds_other
                microseconds = dt.microsecond - other.microsecond
                return timedelta(days, seconds, microseconds)
            elif (isinstance(other, datetime_python) or
                  isinstance(other, cf.real_datetime)):
                # datetime - real_datetime
                if not dt.datetime_compatible:
                    msg = ("Cannot compute the time difference between dates"
                           " with different calendars. One of the datetime"
                           " objects may have been converted to a native"
                           " python datetime instance. Try using"
                           " only_use_cftime_datetimes=True when creating the"
                           " datetime object.")
                    raise TypeError(msg)
                return dt._to_real_datetime() - other
            elif isinstance(other, timedelta):
                # datetime - timedelta
                if dt.calendar == 'decimal360':
                    (year, month, day, hour, minute, second, microsecond,
                     _1, _2) = cf.add_timedelta_360_day(dt, -other)
                else:
                    year, month, day, hour, minute, second, microsecond = (
                        _add_timedelta(dt, -other))
                return datetime(year, month, day,
                                hour, minute, second, microsecond,
                                calendar=dt.calendar,
                                has_year_zero=dt.has_year_zero)
            else:
                return NotImplemented
        else:
            if (isinstance(self, datetime_python) or
                isinstance(self, cf.real_datetime)):
                # real_datetime - datetime
                if not other.datetime_compatible:
                    msg = ("Cannot compute the time difference between dates"
                           " with different calendars. One of the datetime"
                           " objects may have been converted to a native"
                           " python datetime instance. Try using"
                           " only_use_cftime_datetimes=True when creating the"
                           " datetime object.")
                    raise TypeError(msg)
                return self - other._to_real_datetime()
            else:
                return NotImplemented
